import os
import platform
import pprint
import threading
import time
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import dearpygui.dearpygui as dpg

from goofi.data import DataType
from goofi.gui import events
from goofi.gui.data_viewer import ViewerContainer
from goofi.message import Message, MessageType
from goofi.node_helpers import NodeRef, list_nodes
from goofi.params import BoolParam, FloatParam, IntParam, Param, StringParam

DTYPE_SHAPE_MAP = {
    DataType.ARRAY: dpg.mvNode_PinShape_CircleFilled,
    DataType.STRING: dpg.mvNode_PinShape_TriangleFilled,
    DataType.TABLE: dpg.mvNode_PinShape_QuadFilled,
}


NODE_CAT_COLORS = [
    [110, 88, 208],
    [129, 178, 154],
    [242, 204, 143],
    [102, 168, 212],
    [224, 122, 95],
    [178, 144, 205],
]

PARAM_WINDOW_WIDTH = 400


class MetadataPrinter(pprint.PrettyPrinter):
    def format(self, object, context, maxlevels, level):
        if isinstance(object, float):
            return (f"{object:.1f}", True, False)
        return super().format(object, context, maxlevels, level)


data = {
    "apple": 4.123456,
    "banana": 3.987654,
    "fruits": [{"name": "cherry", "value": 2.1234}, {"name": "blueberry", "value": 1.456789}],
}


def format_name(name: str) -> str:
    """Format a name to be used as a node label."""
    return name.replace("_", " ").title()


def running(func):
    """Decorator to make sure the window is running before calling a function."""

    def wrapper(*args, **kwargs):
        # make sure DearPyGui is running
        n_tries = 5
        while not dpg.is_dearpygui_running() and n_tries > 0:
            time.sleep(0.1)
            n_tries -= 1
        if n_tries == 0:
            raise RuntimeError("DearPyGui is not running.")
        return func(*args, **kwargs)

    return wrapper


def update_minimap(func):
    """Decorator to update the minimap state after the node editor is updated."""

    def wrapper(*args, **kwargs):
        func(*args, **kwargs)

        # update minimap state
        if len(args[0].nodes) > 0:
            dpg.configure_item(args[0].node_editor, minimap=True)
        else:
            dpg.configure_item(args[0].node_editor, minimap=False)

    return wrapper


@dataclass
class GUINode:
    item: int
    input_slots: Dict[str, int]
    output_slots: Dict[str, int]
    output_draw_handlers: Dict[str, int]
    node_ref: NodeRef
    _error_msg: str = None

    def __post_init__(self):
        self.set_error(None, Window())

    def set_error(self, msg: str, win: "Window") -> None:
        """Set the error message."""
        if msg is None:
            # no error, reset theme
            try:
                dpg.bind_item_theme(self.item, win.node_themes[self.node_ref.category])
            except SystemError:
                # the node might have been deleted, ignore this error
                pass
            return

        if msg == self._error_msg:
            # nothing changed
            return

        self._error_msg = msg

        # we have an error, set theme
        dpg.bind_item_theme(self.item, win.node_error_themes[self.node_ref.category])

        # update the info window if it is open
        info_win = win.node_info_window
        if info_win is not None and dpg.get_item_user_data(info_win) == self.item:
            dpg.set_value(dpg.get_item_children(info_win)[1][2], msg)

    def display_info(self, win: "Window") -> None:
        """Display information about the node."""
        if win.node_info_window is not None:
            # close the current info window
            dpg.delete_item(win.node_info_window)
            win.node_info_window = None

        # create new info window
        size = (500, 300)
        pos = dpg.get_mouse_pos(local=False)
        pos = (pos[0] - size[0] / 2, pos[1] - size[1] / 2)

        with dpg.theme() as win_theme:
            with dpg.theme_component():
                dpg.add_theme_color(dpg.mvThemeCol_WindowBg, [200, 200, 200, 255], category=dpg.mvThemeCat_Core)
                dpg.add_theme_color(dpg.mvThemeCol_Text, [0, 0, 0, 255], category=dpg.mvThemeCat_Core)
                dpg.add_theme_color(dpg.mvThemeCol_FrameBg, [0, 0, 0, 0], category=dpg.mvThemeCat_Core)

        with dpg.window(
            width=size[0],
            height=size[1],
            pos=pos,
            no_move=True,
            no_close=True,
            no_collapse=True,
            no_resize=True,
            no_title_bar=True,
            user_data=self.item,
        ) as win.node_info_window:
            dpg.add_text("Node Info")
            dpg.add_separator()

            txt = "All good!" if self._error_msg is None else self._error_msg
            dpg.add_input_text(multiline=True, default_value=txt.strip(), readonly=True, width=-1, height=-1)

            dpg.bind_item_theme(win.node_info_window, win_theme)


def handle_data(win: "Window", gui_node: GUINode, node: NodeRef, message: Message):
    """
    Handle a data message from a node. This function is registered as a message handler for the
    `MessageType.DATA` message type.

    ### Parameters
    `gui_node` : GUINode
        The GUI node instance.
    `node` : NodeRef
        The node reference.
    `message` : Message
        The data message.
    """
    try:
        gui_node.output_draw_handlers[message.content["slot_name"]](message)
        gui_node.set_error(None, win)
    except ValueError as e:
        # TODO: add proper logging
        print(f"Output draw handler for slot {message.content['slot_name']} failed: {e}")
    except KeyError:
        print(gui_node, message.content["slot_name"], gui_node.output_draw_handlers.keys())

    if win.metadata_view is not None and win.selected_node == gui_node.item:
        try:
            dpg.set_value(
                win.metadata_view[message.content["slot_name"]],
                MetadataPrinter(compact=True, width=50).pformat(message.content["data"].meta),
            )
        except SystemError:
            # param window was closed, ignore this error
            pass


def toggle_log_plot(_1, _2, data):
    """Toggle the log plot state of a data viewer."""
    _, node, slot, axis = data
    node.output_draw_handlers[slot].toggle_log_plot(axis)


def param_updated(a, value, user_data):
    """
    A GUI callback function for updating a parameter value. This is called by DearPyGui when
    a parameter is updated, and passed the update to the node.
    """
    group, name, node = user_data[:3]

    if len(user_data) == 4:
        # callback from a trigger button, value should be True
        value = user_data[3]
    elif len(user_data) == 5:
        # the parameter includes multiple input widgets, update all of them
        input_group, value_type = user_data[3:]

        try:
            # make sure the value has the correct type
            value = value_type(value)
        except ValueError:
            # failed to convert value to correct type, ignore this update
            return

        # update all input widgets
        for child in dpg.get_item_children(input_group)[1]:
            try:
                dpg.set_value(child, value)
            except SystemError:
                # the input widget might have been deleted, ignore this error
                pass

    # send the updated parameter to the node
    node.update_param(group, name, value)
    # mark manager state as dirty
    Window().manager.unsaved_changes = True


def add_param(parent: int, group: str, name: str, param: Param, node: NodeRef) -> None:
    """
    Add a parameter to the GUI.

    ### Parameters
    `parent` : int
        The parent item.
    `group` : str
        The parameter group.
    `name` : str
        The parameter name.
    `param` : Param
        The parameter.
    `node` : NodeRef
        The node reference.
    """
    with dpg.table_row(parent=parent):
        with dpg.table_cell():
            lbl = dpg.add_text(format_name(name))
            if param.doc is not None:
                dpg.add_text(param.doc, parent=dpg.add_tooltip(parent=lbl))

        with dpg.table_cell():
            if isinstance(param, BoolParam):
                # parameter is a bool
                if param.trigger:
                    dpg.add_button(label=format_name(name), callback=param_updated, user_data=(group, name, node, True))
                else:
                    dpg.add_checkbox(default_value=param.value, callback=param_updated, user_data=(group, name, node))
            elif isinstance(param, FloatParam):
                with dpg.group(horizontal=True) as input_group:
                    # parameter is a float
                    Window().param_input_fields.append(
                        dpg.add_input_text(
                            width=50,
                            scientific=True,
                            default_value=str(param.value),
                            callback=param_updated,
                            user_data=(group, name, node, input_group, float),
                        )
                    )
                    dpg.add_slider_float(
                        default_value=param.value,
                        min_value=param.vmin,
                        max_value=param.vmax,
                        user_data=(group, name, node, input_group, float),
                        callback=param_updated,
                    )
            elif isinstance(param, IntParam):
                with dpg.group(horizontal=True) as input_group:
                    # parameter is an integer
                    Window().param_input_fields.append(
                        dpg.add_input_text(
                            width=50,
                            default_value=str(param.value),
                            callback=param_updated,
                            user_data=(group, name, node, input_group, int),
                        )
                    )
                    dpg.add_slider_int(
                        default_value=param.value,
                        min_value=param.vmin,
                        max_value=param.vmax,
                        callback=param_updated,
                        user_data=(group, name, node, input_group, int),
                    )
            elif isinstance(param, StringParam):
                # parameter is a string
                if param.options is None:
                    # `options` is not set, use an unconstrained text input
                    Window().param_input_fields.append(
                        dpg.add_input_text(
                            default_value=param.value,
                            callback=param_updated,
                            user_data=(group, name, node),
                        )
                    )
                else:
                    # `options` is set, use a dropdown menu
                    dpg.add_combo(
                        default_value=param.value,
                        items=param.options,
                        callback=param_updated,
                        user_data=(group, name, node),
                    )
            else:
                raise NotImplementedError(f"Parameter type {type(param)} not implemented.")


def add_output_slot(parent: int, name: str, closed: bool = False, size: Tuple[int, int] = (175, 100)) -> int:
    """
    Add an output slot to the GUI, which consists of a collapsable header and a content window for the data viewer.

    ### Parameters
    `parent` : int
        The parent item (the node attribute).
    `name` : str
        The name of the output slot.
    `closed` : bool
        Whether the headers should be closed initially.
    `size` : Tuple[int, int]
        The size of the content area.

    ### Returns
    `content` : int
        The content window item.
    """

    def header_callback(_1, _2, items: List[int]):
        """
        Event callback for the collapsable header, both for general clicks, and for toggle_open events.
        This is a workaround for the fact that toggle_close events don't exist. As toggle_open is always
        called after click, clicking on the header always closes it, so the toggle_open callback is used
        reopen it if the internal header state is open.

        ### Parameters
        `items` : List[int]
            A list of four items: a bool indicating to close the header, the header, the window, and the content window.
        """
        close, header, window, content = items
        header_height = dpg.get_item_state(header)["rect_size"][1]

        if header_height == 0:
            # this happens when the header is off screen, set to default height
            header_height = 15

        if close:
            # header was closed, shrink window to header size
            dpg.set_item_height(window, header_height)
        else:
            # header was opened, expand window to header size plus cotent size
            content_height = dpg.get_item_state(content)["rect_size"][1]
            if content_height == 0:
                # this happens when the content window is off screen, set to default height
                content_height = size[1]

            dpg.set_item_height(window, header_height + content_height + 4)  # magic + 4 to avoid scroll bar

    # NOTE: The user_data lists are used to pass data to the callbacks. The first element is a bool, which
    # is used to indicate whether the header was closed or opened. The remaining elements are the header,
    # window, and content window items. Due to the way DearPyGui registers callbacks we can not use partials here.
    user_data_open, user_data_close = [False], [True]
    # register handlers
    with dpg.item_handler_registry() as reg:
        dpg.add_item_clicked_handler(callback=header_callback, user_data=user_data_close)
        dpg.add_item_toggled_open_handler(callback=header_callback, user_data=user_data_open)

    # NOTE: we initially set the height to 1 to avoid the node reaching out of the window, which would break the
    # initial header_callback call as the header and content sizes according to DearPyGui would be 0.
    # The child window is needed, because otherwise the header will expand to fill the entire node editor window.
    with dpg.child_window(width=size[0], height=1, parent=parent) as window:
        # add collapsable content area for the data viewer
        header = dpg.add_collapsing_header(label=name, default_open=not closed, open_on_arrow=False, closable=False)
        content = dpg.add_child_window(width=size[0], height=size[1], user_data=size, parent=header)
        with dpg.tooltip(parent=content, show=True):
            dpg.add_text("- ctrl + click\nswitch between data viewers\n- scroll\nzoom in/out")

    # store header, window and content window in user_data for the two callbacks
    user_data_open.extend([header, window, content])
    user_data_close.extend([header, window, content])
    # bind the two callbacks to the header
    dpg.bind_item_handler_registry(header, reg)

    # schedule the header callback to set up the window sizes
    initial_args = user_data_close if closed else user_data_open
    threading.Timer(0.1, header_callback, [None, None, initial_args]).start()
    return content


class Window:
    """
    The graphical user interface window, implemented as a threaded singleton. The window is created
    when Window() is first called. Subsequent calls to Window() will return the same instance. The
    window thread is a daemon thread, so it will be terminated when the main thread exits, or when
    Window().close() is called.

    ### Parameters
    `manager` : Manager
        The manager instance.
    """

    _instance = None
    _initialized = False

    def __new__(cls, manager=None):
        if cls._instance is None:
            # make sure the window is created in the main thread (required by MacOS)
            if threading.current_thread() != threading.main_thread():
                raise RuntimeError("Window must be created in the main thread.")

            # instantiate the window
            # TODO: add proper logging
            print("Starting graphical user interface.")
            cls._instance = super(Window, cls).__new__(cls)
            cls._instance._initialize(manager)
        return cls._instance

    @running
    @update_minimap
    def add_node(
        self,
        node_name: str,
        node: NodeRef,
        pos: Optional[Tuple[int, int]] = None,
        offset: Optional[Tuple[int, int]] = None,
        viewers: Optional[List[Dict[str, Any]]] = None,
    ) -> None:
        """
        Outward facing method to add a node to the GUI.

        ### Parameters
        `node_name` : str
            The name of the node.
        `node` : NodeRef
            The node reference.
        `pos` : Optional[Tuple[int, int]]
            Position of the node. If None, uses the mouse position.
        `offset` : Optional[Tuple[int, int]]
            Offsets the node position by the given amount.
        `viewers` : Optional[List[Dict[str, Any]]]
            List of data viewer configurations.
        """
        if pos is None:
            pos = self._node_editor_mouse_pos()
        if offset is not None:
            pos = (pos[0] + offset[0], pos[1] + offset[1])

        with dpg.node(parent=self.node_editor, label=node_name, pos=pos, user_data=node) as node_id:
            ############### input slots ###############
            in_slots = {}
            for name, dtype in node.input_slots.items():
                # create input slot
                in_slots[name] = dpg.add_node_attribute(
                    label=name, attribute_type=dpg.mvNode_Attr_Input, shape=DTYPE_SHAPE_MAP[dtype], user_data=dtype
                )
                # simply add a text label
                dpg.add_text(name, parent=in_slots[name])

            ############### output slots ###############
            out_slots = {}
            output_draw_handlers = {}
            for name, dtype in node.output_slots.items():
                # create output slot
                out_slots[name] = dpg.add_node_attribute(
                    label=name, attribute_type=dpg.mvNode_Attr_Output, shape=DTYPE_SHAPE_MAP[dtype], user_data=dtype
                )

                # determine data viewer configuration
                viewer_kwargs = viewers[name] if viewers is not None and name in viewers else {}
                collapsed = viewer_kwargs.pop("collapsed", len(node.output_slots) > 2)

                # create content window for data viewer (initialize closed if more than two output slots)
                content = add_output_slot(out_slots[name], name, closed=collapsed)
                # create data viewer
                output_draw_handlers[name] = ViewerContainer(dtype, content, **viewer_kwargs)

            # add node to node list
            self.nodes[node_name] = GUINode(node_id, in_slots, out_slots, output_draw_handlers, node)

            # TODO: register PROCESSING_ERROR message handler and display error messages
            node.set_message_handler(
                MessageType.PROCESSING_ERROR, partial(self._processing_error_callback, node_name=node_name)
            )

            # register data message handler to update the data viewers
            node.set_message_handler(MessageType.DATA, partial(handle_data, self, self.nodes[node_name]))

    @running
    def remove_node(self, name: str) -> None:
        """
        Outward facing method to remove a node from the GUI.

        ### Parameters
        `name` : str
            The name of the node.
        """
        self._remove_node(self.nodes[name], notify_manager=False)

    @running
    @update_minimap
    def _remove_node(self, item: int, notify_manager: bool = True) -> None:
        """
        Remove a node from the GUI.

        ### Parameters
        `item` : int
            The node item.
        `notify_manager` : bool
            Whether to notify the manager to remove the node.
        """
        if item in dpg.get_selected_nodes(self.node_editor):
            # deselect node if it is selected
            self._select_node(None)

        # determine the node name
        name = dpg.get_item_label(item)

        # remove links associated with node
        for link in list(self.links.keys()):
            if link[0] == name or link[1] == name:
                self._remove_link(self.links[link], notify_manager=False)

        # unregister node by deleting it from the node list
        del self.nodes[name]

        # delete node from gui
        dpg.delete_item(item)

        if notify_manager:
            # remove node from manager
            self.manager.remove_node(name, notify_gui=False)

    @running
    def add_link(self, node_out: str, node_in: str, slot_out: str, slot_in: str) -> None:
        """
        Outward facing method to add a link between two nodes.

        ### Parameters
        `node_out` : str
            The name of the output node.
        `node_in` : str
            The name of the input node.
        `slot_out` : str
            The output slot name of `node_out`.
        `slot_in` : str
            The input slot name of `node_in`.
        """
        # determine slot IDs
        slot1 = self.nodes[node_out].output_slots[slot_out]
        slot2 = self.nodes[node_in].input_slots[slot_in]
        # add link internally
        self._add_link(self.node_editor, (slot1, slot2), notify_manager=False)

    @running
    def _add_link(self, sender: int, items: Tuple[int, int], notify_manager: bool = True) -> None:
        """
        Add a link between two nodes to the GUI.

        ### Parameters
        `sender` : int
            The sender item.
        `items` : Tuple[int, int]
            The two items to link.
        `notify_manager` : bool
            Whether to notify the manager to add the link.
        """
        # get slot names
        slot1 = dpg.get_item_label(items[0])
        slot2 = dpg.get_item_label(items[1])
        # get node names
        node1 = dpg.get_item_label(dpg.get_item_parent(items[0]))
        node2 = dpg.get_item_label(dpg.get_item_parent(items[1]))

        # make sure the data types match
        dtype1 = dpg.get_item_user_data(items[0])
        dtype2 = dpg.get_item_user_data(items[1])
        if dtype1 != dtype2:
            # TODO: add proper logging
            print(f"Data types of slots {slot1} ({dtype1.name}) and {slot2} ({dtype2.name}) do not match.")
            return

        # remove link first if the input slot already has a link
        for link in list(self.links.keys()):
            if link[1] == node2 and link[3] == slot2:
                self._remove_link(self.links[link], notify_manager=True)

        # register link
        self.links[(node1, node2, slot1, slot2)] = dpg.add_node_link(items[0], items[1], parent=sender)

        if notify_manager:
            # add link to manager
            self.manager.add_link(node1, node2, slot1, slot2, notify_gui=False)

    def remove_link(self, node_out: str, node_in: str, slot_out: str, slot_in: str) -> None:
        """
        Outward facing method to remove a link between two nodes.

        ### Parameters
        `node_out` : str
            The name of the output node.
        `node_in` : str
            The name of the input node.
        `slot_out` : str
            The output slot name of `node_out`.
        `slot_in` : str
            The input slot name of `node_in`.
        """
        self._remove_link(self.links[(node_out, node_in, slot_out, slot_in)], notify_manager=False)

    @running
    def _remove_link(self, item: int, notify_manager: bool = True) -> None:
        """
        Remove a link between two nodes from the GUI.

        ### Parameters
        `item` : int
            The link item.
        `notify_manager` : bool
            Whether to notify the manager to remove the link.
        """
        conf = dpg.get_item_configuration(item)
        # get slot names
        slot1, slot2 = dpg.get_item_label(conf["attr_1"]), dpg.get_item_label(conf["attr_2"])
        # get node names
        node1 = dpg.get_item_label(dpg.get_item_parent(conf["attr_1"]))
        node2 = dpg.get_item_label(dpg.get_item_parent(conf["attr_2"]))

        if notify_manager:
            # remove link from manager
            self.manager.remove_link(node1, node2, slot1, slot2, notify_gui=False)

        # unregister link
        del self.links[(node1, node2, slot1, slot2)]
        # delete link from gui
        dpg.delete_item(item)

    def load(self, path: Optional[str] = None) -> None:
        """
        Load a goofi-pipe from a file.

        ### Parameters
        `path` : Optional[str]
            Optional path to load the GUI state from. If None, opens a file dialog.
        """

        def file_callback(_, info):
            self.load(path=info["file_path_name"])

        if path is None:
            # no save path set, open save dialog
            self._get_file(file_callback, message="Select a file to load from.")
            return

        try:
            # try loading the manager
            self.manager.load(filepath=path)
        except FileNotFoundError:
            # file does not exist, open error dialog
            w, h = dpg.get_viewport_width(), dpg.get_viewport_height()
            pos = (w / 2 - w / 8, h / 2 - h / 12)
            with dpg.window(label="Error", width=w / 4, height=h / 6, pos=pos) as win:
                dpg.add_text("The file does not exist.")

                dpg.add_separator()
                dpg.add_text(path)
                dpg.add_separator()

                # add button
                dpg.add_button(label="Ok", callback=lambda _, __: dpg.delete_item(win))
        except RuntimeError:
            # file does not exist, open error dialog
            w, h = dpg.get_viewport_width(), dpg.get_viewport_height()
            pos = (w / 2 - w / 8, h / 2 - h / 12)
            with dpg.window(label="Error", width=w / 4, height=h / 6, pos=pos) as win:
                dpg.add_text("The current pipeline is not empty.")

                # add button
                dpg.add_button(label="Ok", callback=lambda _, __: dpg.delete_item(win))

    def save(
        self,
        path: Optional[str] = None,
        overwrite: bool = True,
        save_as: bool = False,
        callback: Optional[Callable] = None,
    ) -> None:
        """
        Save the current state of the GUI.

        ### Parameters
        `path` : Optional[str]
            Optional path to save the GUI state to. If None, uses the current save path.
        `overwrite` : bool
            Whether to overwrite the current save path.
        `save_as` : bool
            Whether to open a file dialog to select a save path.
        `callback` : Optional[Callable]
            Optional callback function to call when the save is complete.
        """

        def file_callback(_, info):
            # save the manager to the selected path
            self.save(path=info["file_path_name"], overwrite=False, callback=callback)

        if save_as or (self.manager.save_path is None and path is None):
            # no save path set, open save dialog
            self._get_file(file_callback, message="Select a file to save to.")
            return

        try:
            # try saving the manager
            self.manager.save(filepath=path, overwrite=overwrite)

            if callback is not None:
                callback()
        except FileExistsError:

            def confirm_callback(_1, _2, data):
                """Callback for the overwrite confirmation dialog."""
                win, confirm = data
                if confirm:
                    self.save(path=path, overwrite=True, callback=callback)
                # close the confirmation dialog
                dpg.delete_item(win)

            # file already exists, open overwrite dialog
            w, h = dpg.get_viewport_width(), dpg.get_viewport_height()
            pos = (w / 2 - w / 8, h / 2 - h / 12)
            with dpg.window(label="Overwrite?", width=w / 4, height=h / 6, pos=pos) as win:
                dpg.add_text("The file already exists. Overwrite?")

                if path is not None:
                    dpg.add_separator()
                    dpg.add_text(path)
                    dpg.add_separator()

                # add buttons
                with dpg.group(horizontal=True):
                    dpg.add_button(label="Yes", callback=confirm_callback, user_data=(win, True))
                    dpg.add_button(label="Cancel", callback=confirm_callback, user_data=(win, False))
        except RuntimeError as e:
            # something went wrong, open error dialog
            w, h = dpg.get_viewport_width(), dpg.get_viewport_height()
            pos = (w / 2 - w / 8, h / 2 - h / 12)
            with dpg.window(label="Error", width=w / 4, height=h / 6, pos=pos) as win:
                dpg.add_text(str(e))

                # add button
                dpg.add_button(label="Ok", callback=lambda _, __: dpg.delete_item(win))

    def _get_file(self, callback: Callable, message: Optional[str] = None) -> None:
        """
        Open a file dialog (e.g. for loading or saving a file).

        ### Parameters
        `callback` : Callable
            The callback function to call when the file is selected.
        `message` : Optional[str]
            Optional message to display in the save dialog.
        """

        # open file selection dialog
        width, height = dpg.get_viewport_width() / 2, dpg.get_viewport_height() / 2
        with dpg.file_dialog(
            label=message,
            callback=callback,
            default_path=".",
            default_filename="",
            modal=True,
            width=width,
            height=height,
        ) as self.file_selection_window:
            dpg.add_file_extension(".gfi", label="goofi-pipes")
            dpg.add_file_extension(".*", label="All Files")

    def get_node_state(self, name: str) -> Dict[str, Any]:
        """
        Get the state of a node.

        ### Parameters
        `name` : str
            The name of the node.

        ### Returns
        `state` : Dict[str, Any]
            The state of the node.
        """
        if name not in self.nodes:
            # TODO: add proper logging
            print(f"Can't get state of node {name}. The node does not exist.")
            return None

        node = self.nodes[name]
        return {
            "pos": dpg.get_item_pos(node.item),
            "viewers": {slot: viewer.get_state() for slot, viewer in node.output_draw_handlers.items()},
        }

    def _select_node(self, item: Optional[int]) -> None:
        """
        Register or unregister a node as selected. If a new node is selected, the parameters window is
        updated to show the parameters of that node.

        ### Parameters
        `item` : Optional[int]
            The node item. If None, deselects the currently selected node.
        """
        if self.selected_node == item:
            # do nothing if the node is already selected
            return

        # update window layout
        self.selected_node = item
        self.resize_callback(None, [dpg.get_viewport_width(), dpg.get_viewport_height()])
        # clear parameters window
        dpg.delete_item(self.side_panel_win, children_only=True)

        if item is None or dpg.get_item_label(item) not in self.nodes:
            self.selected_node = None
            # node deselected, hide parameters window and resize node editor
            dpg.configure_item(self.side_panel_win, show=False)
            self.resize_callback(None, [dpg.get_viewport_width(), dpg.get_viewport_height()])
            return

        # get node reference
        node = self.nodes[dpg.get_item_label(item)]
        node_ref = node.node_ref

        with dpg.child_window(height=dpg.get_viewport_height() / 2, parent=self.side_panel_win):
            # add title
            dpg.add_text("Parameters")
            dpg.add_separator()

            # populate parameters window
            self.param_input_fields.clear()
            with dpg.tab_bar():
                for group in node_ref.params:
                    with dpg.tab(label=format_name(group)) as tab:
                        with dpg.table(header_row=False, parent=tab, policy=dpg.mvTable_SizingStretchProp) as table:
                            dpg.add_table_column()
                            dpg.add_table_column()

                            for name, param in node_ref.params[group].items():
                                add_param(table, group, name, param, node_ref)

        with dpg.child_window(autosize_y=True, parent=self.side_panel_win):
            # add title
            dpg.add_text("Outputs")
            dpg.add_separator()

            # add one metadata view for each output slot
            self.metadata_view = {}
            with dpg.tab_bar():
                for slot in node_ref.output_slots:
                    with dpg.tab(label=slot) as tab:
                        with dpg.group(horizontal=True):
                            dpg.add_checkbox(
                                label="log-scale x-axis",
                                default_value=node.output_draw_handlers[slot].log_scale_x,
                                callback=toggle_log_plot,
                                user_data=(self, node, slot, "x"),
                            )
                            dpg.add_checkbox(
                                label="log-scale y-axis",
                                default_value=node.output_draw_handlers[slot].log_scale_y,
                                callback=toggle_log_plot,
                                user_data=(self, node, slot, "y"),
                            )
                        dpg.add_separator()
                        self.metadata_view[slot] = dpg.add_input_text(
                            default_value="", multiline=True, readonly=True, width=-1, height=-1
                        )

        # show parameters window
        dpg.configure_item(self.side_panel_win, show=True)

    def _processing_error_callback(self, node: NodeRef, message: Message, node_name: str) -> None:
        """Callback for the `MessageType.PROCESSING_ERROR` message type."""
        error = message.content["error"]
        if error is None:
            # no error, clear error message
            self.nodes[node_name].set_error(None, self)
            return

        self.nodes[node_name].set_error(error, self)
        print(f"Error in node {node_name}:\n{error}")

    def link_callback(self, sender: int, items: Tuple[int, int]) -> None:
        """Callback from DearPyGui that two nodes were connected."""
        self._add_link(sender, items)

    def delink_callback(self, _, item: int) -> None:
        """Callback from DearPyGui that a link was removed."""
        self._remove_link(item)

    def resize_callback(self, _, size: Tuple[int, int]) -> None:
        """Callback from DearPyGui that the viewport was resized."""
        # resize window to fill viewport
        if platform.system() == "Windows":
            # frame border seems to be different on Windows
            dpg.configure_item(self.window, width=size[0] - 16, height=size[1] - 39)
        else:
            dpg.configure_item(self.window, width=size[0], height=size[1])

        if self.selected_node is None:
            # no node selected, resize node editor to fill viewport
            dpg.configure_item(self.node_editor, width=0)
        else:
            # node selected, resize node editor to fill viewport minus parameters window
            dpg.configure_item(self.node_editor, width=size[0] - PARAM_WINDOW_WIDTH)

            children = dpg.get_item_children(self.side_panel_win)[1]
            if len(children) > 0:
                dpg.set_item_height(children[0], dpg.get_viewport_height() / 2)

    def update_title(self) -> None:
        """Update the window title."""
        start = time.time()
        while not hasattr(self, "manager"):
            # wait until the manager is initialized
            if time.time() - start > 1:
                raise RuntimeError("Timeout while waiting for manager to initialize.")
            time.sleep(0.01)

        # update window title
        unsaved_changes = "*" if self.manager.unsaved_changes else ""
        if self.manager.save_path is None:
            dpg.set_viewport_title(f"{unsaved_changes}goofi-pipe")
        else:
            dpg.set_viewport_title(f"goofi-pipe | {unsaved_changes}{self.manager.save_path.split(os.sep)[-1]}")

    def exit_callback(self, value):
        """Callback from DearPyGui that the close button was pressed."""
        if self.manager.unsaved_changes and not self.unsaved_changes_dialog_open:

            def callback(_1, _2, data):
                """Callback for the unsaved changes dialog."""
                win, save = data
                # close the confirmation dialog
                dpg.delete_item(win)
                self.unsaved_changes_dialog_open = False

                if save:
                    # save changes first
                    self.save(callback=self.terminate)
                else:
                    # close without saving
                    self.terminate()

            # unsaved changes, open confirm dialog
            self.unsaved_changes_dialog_open = True
            w, h = dpg.get_viewport_width(), dpg.get_viewport_height()
            pos = (w / 2 - w / 8, h / 2 - h / 12)
            with dpg.window(label="Save before closing?", width=w / 4, height=h / 6, pos=pos) as win:
                dpg.add_text("There are unsaved changes.\nDo you want to save before closing?")

                # add buttons
                with dpg.group(horizontal=True):
                    dpg.add_button(label="Save", callback=callback, user_data=(win, True))
                    dpg.add_button(label="Close without saving", callback=callback, user_data=(win, False))
            return

        # no unsaved changes, close window
        self.terminate()

    def _to_node_editor_coords(self, pos: Tuple[int, int]) -> Tuple[int, int]:
        """Convert a position from window space to node editor space."""
        # get reference node position
        orig_ref_pos = dpg.get_item_user_data("_ref")
        curr_ref_pos = dpg.get_item_state("_ref")["rect_min"]
        # calculate mouse position within node editor
        return pos[0] - curr_ref_pos[0] + orig_ref_pos[0], pos[1] - curr_ref_pos[1] + orig_ref_pos[1]

    def _to_window_coords(self, pos: Tuple[int, int]) -> Tuple[int, int]:
        """Convert a position from node editor space to window space."""
        # get reference node position
        orig_ref_pos = dpg.get_item_user_data("_ref")
        curr_ref_pos = dpg.get_item_state("_ref")["rect_min"]
        # calculate mouse position within node editor
        return pos[0] - curr_ref_pos[0] + orig_ref_pos[0], pos[1] - curr_ref_pos[1] + orig_ref_pos[1]

    def _node_editor_mouse_pos(self) -> Tuple[int, int]:
        """Get the mouse position within the node editor."""
        return self._to_node_editor_coords(dpg.get_mouse_pos(local=False))

    def _register_node_category_themes(self, darkness: float = 0.5) -> None:
        """Register themes for each node category."""
        cats = [n.category() for n in list_nodes()]
        cats = sorted(list(set(cats)))

        def scale(color, s):
            return [max(min(int(c * s), 255), 0) for c in color]

        # set up node themes
        self.node_themes = {}
        for i, cat in enumerate(cats):
            with dpg.theme() as theme:
                with dpg.theme_component():
                    dpg.add_theme_color(
                        dpg.mvNodeCol_TitleBar, scale(NODE_CAT_COLORS[i], darkness), category=dpg.mvThemeCat_Nodes
                    )
                    dpg.add_theme_color(
                        dpg.mvNodeCol_TitleBarHovered, scale(NODE_CAT_COLORS[i], darkness + 0.1), category=dpg.mvThemeCat_Nodes
                    )
                    dpg.add_theme_color(
                        dpg.mvNodeCol_TitleBarSelected, scale(NODE_CAT_COLORS[i], darkness + 0.2), category=dpg.mvThemeCat_Nodes
                    )

            self.node_themes[cat] = theme

        # set up node error themes
        self.node_error_themes = {}
        for i, cat in enumerate(cats):
            with dpg.theme() as theme:
                with dpg.theme_component():
                    dpg.add_theme_color(dpg.mvNodeCol_NodeBackground, [153, 20, 22], category=dpg.mvThemeCat_Nodes)
                    dpg.add_theme_color(dpg.mvNodeCol_NodeBackgroundHovered, [173, 20, 22], category=dpg.mvThemeCat_Nodes)
                    dpg.add_theme_color(dpg.mvNodeCol_NodeBackgroundSelected, [193, 20, 22], category=dpg.mvThemeCat_Nodes)

                    dpg.add_theme_color(
                        dpg.mvNodeCol_TitleBar, scale(NODE_CAT_COLORS[i], darkness), category=dpg.mvThemeCat_Nodes
                    )
                    dpg.add_theme_color(
                        dpg.mvNodeCol_TitleBarHovered, scale(NODE_CAT_COLORS[i], darkness + 0.1), category=dpg.mvThemeCat_Nodes
                    )
                    dpg.add_theme_color(
                        dpg.mvNodeCol_TitleBarSelected, scale(NODE_CAT_COLORS[i], darkness + 0.2), category=dpg.mvThemeCat_Nodes
                    )
            self.node_error_themes[cat] = theme

    def _initialize(self, manager, width=1280, height=720):
        """Initialize the window and launch the event loop (blocking)."""
        dpg.create_context()
        self.manager = manager

        # initialize dicts to map names to dpg items
        self.nodes = {}
        self.links = {}
        self.selected_node = None
        self.create_node_window = None
        self.last_create_node_tab = 0
        self.file_selection_window = None
        self.unsaved_changes_dialog_open = False
        self.node_clipboard = None
        self.node_info_window = None
        self.metadata_view = None
        self.param_input_fields = []

        # create window
        self.window = dpg.add_window(
            label="window",
            width=width,
            height=height,
            no_title_bar=True,
            no_resize=True,
            no_move=True,
            no_close=True,
            no_collapse=True,
        )

        # create menu bar
        with dpg.menu_bar(parent=self.window):
            dpg.add_menu_item(label="Load", callback=lambda: self.load())
            dpg.add_menu_item(label="Save As", callback=lambda: self.save(save_as=True))
            dpg.add_menu_item(label="Save", callback=lambda: self.save())

        with dpg.group(horizontal=True, parent=self.window):
            # create node editor
            self.node_editor = dpg.add_node_editor(callback=self.link_callback, delink_callback=self.delink_callback)
            # add parameters window
            self.side_panel_win = dpg.add_child_window(label="Parameters", autosize_x=True, border=False)

        self._register_node_category_themes()

        # hide the reference node using themes
        with dpg.theme() as ref_theme:
            with dpg.theme_component():
                dpg.add_theme_color(dpg.mvNodeCol_TitleBar, [0, 0, 0, 0], category=dpg.mvThemeCat_Nodes)
                dpg.add_theme_color(dpg.mvNodeCol_TitleBarSelected, [0, 0, 0, 0], category=dpg.mvThemeCat_Nodes)
                dpg.add_theme_color(dpg.mvNodeCol_TitleBarHovered, [0, 0, 0, 0], category=dpg.mvThemeCat_Nodes)
                dpg.add_theme_color(dpg.mvNodeCol_NodeBackground, [0, 0, 0, 0], category=dpg.mvThemeCat_Nodes)
                dpg.add_theme_color(dpg.mvNodeCol_NodeBackgroundHovered, [0, 0, 0, 0], category=dpg.mvThemeCat_Nodes)
                dpg.add_theme_color(dpg.mvNodeCol_NodeBackgroundSelected, [0, 0, 0, 0], category=dpg.mvThemeCat_Nodes)
                dpg.add_theme_color(dpg.mvNodeCol_NodeOutline, [0, 0, 0, 0], category=dpg.mvThemeCat_Nodes)
                dpg.add_theme_style(dpg.mvNodeStyleVar_NodePadding, 0, 0, category=dpg.mvThemeCat_Nodes)

        # create a reference node to calculate the mouse position within the node editor
        # NOTE: this is a workaround as DearPyGui doesn't provide access to the node editor coordinates
        # NOTE: ideally we would set show=False in the ref node, but this causes a Segmentation Fault
        # NOTE: show=False is now fixed (https://github.com/hoffstadt/DearPyGui/pull/2225), waiting for release
        pos = [0, 0]
        dpg.add_node(label=" ", tag="_ref", parent=self.node_editor, pos=pos, user_data=pos, draggable=False)
        dpg.bind_item_theme("_ref", ref_theme)

        # register user interaction handlers
        with dpg.handler_registry():
            dpg.add_key_release_handler(callback=events.key_release_callback, user_data=self)
            dpg.add_mouse_click_handler(callback=events.click_callback, user_data=self)
            dpg.add_mouse_double_click_handler(callback=events.double_click_callback, user_data=self)

        # register viewport resize handler
        dpg.set_viewport_resize_callback(self.resize_callback)

        # register exit handler
        dpg.set_exit_callback(self.exit_callback)

        # set up window
        self.resize_callback(None, [width, height])

        # start DearPyGui
        dpg.create_viewport(title="goofi-pipe", width=width, height=height, disable_close=True)
        dpg.set_viewport_large_icon("assets/goofi.ico")
        dpg.set_viewport_small_icon("assets/goofi.ico")
        dpg.setup_dearpygui()
        dpg.show_viewport()

        # we are done initializing, start the event loop (blocking)
        self._initialized = True
        dpg.start_dearpygui()

        # cleanup after event loop
        dpg.destroy_context()

        # TODO: add proper logging
        print("Shutting down graphical user interface.")

        # terminate manager
        manager.terminate(notify_gui=False)

    def terminate(self):
        """Stop DearPyGui and close the window."""
        dpg.stop_dearpygui()



import pickle
import queue
import tempfile
import threading
import time
from abc import ABC, abstractmethod
from multiprocessing import Pipe
from multiprocessing.connection import _ConnectionBase
from multiprocessing.managers import BaseManager
from typing import Dict, Tuple, Type

import zmq


class Connection(ABC):
    _CONNECTION_IDS = None
    _BACKEND = None

    def __init__(self) -> None:
        if Connection._CONNECTION_IDS is None:
            raise RuntimeError("Connection._CONNECTION_IDS is None. Call Connection.set_backend() first.")

        # register a unique id for the connection
        self._id = 0
        while self._id in Connection._CONNECTION_IDS:
            self._id += 1
        Connection._CONNECTION_IDS.append(self._id)

    @staticmethod
    def get_backends() -> Dict[str, Type["Connection"]]:
        """
        List all available connection backends.
        """
        return {
            "zmq-tcp": TCPZeroMQConnection,
            "zmq-ipc": IPCZeroMQConnection,
            "mp": MultiprocessingConnection,
        }

    @staticmethod
    def set_backend(backend: str, mp_manager: BaseManager) -> None:
        """
        Set the backend to use for creating connections and initialize a shared set of connection ids.

        ### Parameters
        `backend` : str
            The backend to use. Choose from "zmq-tcp", "zmq-ipc" or "mp".
        `mp_manager` : multiprocessing.Manager
            The multiprocessing manager to use for creating shared objects.
        """
        # initialize a shared set of connection ids
        assert Connection._CONNECTION_IDS is None, "Connection._CONNECTION_IDS is not None."
        Connection._CONNECTION_IDS = mp_manager.list()

        # set the backend if discovering connection IDs was successful (prevents overwriting the backend)
        assert (
            backend in Connection.get_backends().keys()
        ), f"Invalid backend: {backend}. Choose from {list(Connection.get_backends().keys())}"
        Connection._BACKEND = backend

    @staticmethod
    def create() -> Tuple["Connection", "Connection"]:
        """
        Create two instances of the connection class and return them as a tuple. Both instances should be
        connected to each other using the underlying connection mechanism of the deriving class.
        """
        assert Connection._BACKEND is not None, "No backend set. Call Connection.set_backend() first."
        return Connection.get_backends()[Connection._BACKEND]._create()

    @staticmethod
    @abstractmethod
    def _create() -> Tuple["Connection", "Connection"]:
        """
        Create two instances of the connection class and return them as a tuple. Both instances should be
        connected to each other using the underlying connection mechanism of the deriving class.
        """
        raise TypeError("This is an abstract method.")

    @abstractmethod
    def send(self, obj: object) -> None:
        """
        Send an object through the connection (requires the object to be picklable).

        This method should raise a ConnectionError if either it, or the receiving end of the connection is closed.

        ### Parameters
        `obj` : object
            The object to send through the connection.
        """
        pass

    @abstractmethod
    def recv(self) -> object:
        """
        Receive an object through the connection.

        This method should raise a ConnectionError if either it, or the sending end of the connection is closed.

        ### Returns
        `object`
            The object received through the connection.
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close the connection and free any resources associated with it.
        """
        pass

    def try_send(self, obj: object) -> bool:
        """
        Try to send an object through the connection (i.e. catch any ConnectionError that might be raised).

        ### Parameters
        `obj` : object
            The object to send through the connection.

        ### Returns
        `bool`
            True if the object was sent, False otherwise.
        """
        try:
            self.send(obj)
            return True
        except ConnectionError:
            return False

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Connection):
            return False
        return self._id == other._id


class MultiprocessingConnection(Connection):
    def __init__(self, conn: _ConnectionBase) -> None:
        super().__init__()
        self.conn = conn

    @staticmethod
    def _create() -> Tuple[Connection, Connection]:
        conn1, conn2 = Pipe()
        return MultiprocessingConnection(conn1), MultiprocessingConnection(conn2)

    def send(self, obj: object) -> None:
        try:
            self.conn.send(obj)
        except (OSError, BrokenPipeError):
            raise ConnectionError("Connection closed")

    def recv(self) -> object:
        try:
            return self.conn.recv()
        except (OSError, EOFError, TypeError):
            raise ConnectionError("Connection closed")

    def close(self) -> None:
        try:
            self.conn.close()
        except Exception:
            pass

    def __del__(self) -> None:
        """Destructor to close the connection and free any resources associated with it."""
        try:
            self.close()
            self.conn._handle = None
        except Exception:
            pass


class ZeroMQConnection(Connection, ABC):
    def __init__(self, push_endpoint: str, pull_endpoint: str = None) -> None:
        super().__init__()
        self.context = zmq.Context()
        self.push_endpoint = push_endpoint
        self.pull_endpoint = pull_endpoint
        self.push_queue = queue.Queue()
        self.pull_queue = queue.Queue()

        self.alive = True
        self.push_thread = threading.Thread(target=self.run_push, daemon=True)
        self.pull_thread = threading.Thread(target=self.run_pull, daemon=True)

    @classmethod
    def _create(cls) -> Tuple[Connection, Connection]:
        if cls.protocol() == "tcp":
            endpoint1 = TCPZeroMQConnection.unique_tcp_endpoint()
            endpoint2 = TCPZeroMQConnection.unique_tcp_endpoint()
        elif cls.protocol() == "ipc":
            endpoint1 = IPCZeroMQConnection.unique_ipc_endpoint()
            endpoint2 = IPCZeroMQConnection.unique_ipc_endpoint()
        else:
            raise ValueError(f"Invalid protocol: {cls.protocol}")
        return cls(endpoint1, endpoint2), cls(endpoint2, endpoint1)

    def run_push(self):
        push_socket = None
        if self.push_endpoint is not None:
            push_socket = self.context.socket(zmq.PUSH)
            push_socket.connect(self.push_endpoint)

        while self.alive:
            try:
                obj = self.push_queue.get(block=False)
                push_socket.send(pickle.dumps(obj))
            except queue.Empty:
                pass
            time.sleep(0.01)

        if push_socket:
            push_socket.close()

    def run_pull(self):
        if self.pull_endpoint is None:
            # no pull endpoint, stop the thread
            return

        pull_socket = self.context.socket(zmq.PULL)
        pull_socket.bind(self.pull_endpoint)

        while self.alive:
            try:
                obj = pickle.loads(pull_socket.recv(zmq.NOBLOCK))
                self.pull_queue.put(obj)
            except zmq.Again:
                pass
            time.sleep(0.01)

        if pull_socket:
            pull_socket.close()

    def send(self, obj: object) -> None:
        if not self.alive:
            raise ConnectionError("Connection closed")
        if not self.push_thread.is_alive():
            try:
                self.push_thread.start()
            except RuntimeError:
                pass

        self.push_queue.put(obj)

    def recv(self) -> object:
        if not self.alive:
            raise ConnectionError("Connection closed")
        if not self.pull_thread.is_alive():
            try:
                self.pull_thread.start()
            except RuntimeError:
                pass

        return self.pull_queue.get()

    def close(self) -> None:
        self.alive = False
        if self.push_thread.is_alive():
            self.push_thread.join()
        if self.pull_thread.is_alive():
            self.pull_thread.join()

    def __reduce__(self):
        return (self.__class__, (self.push_endpoint, None), self.__getstate__())

    def __getstate__(self):
        return {"_id": self._id}

    def __setstate__(self, state):
        self._id = state["_id"]

    def __del__(self) -> None:
        """Destructor to close the connection and free any resources associated with it."""
        self.close()

    @staticmethod
    @abstractmethod
    def protocol() -> str:
        """Return the protocol used by the connection"""
        pass

    @staticmethod
    def unique_ipc_endpoint() -> str:
        """Generate a unique temporary file path. The file is deleted immediately after creation to avoid clashes"""
        temp_file = tempfile.NamedTemporaryFile(delete=True)
        endpoint = f"ipc://{temp_file.name}"
        # delete the file
        temp_file.close()
        return endpoint

    @staticmethod
    def unique_tcp_endpoint() -> str:
        """Generate a unique TCP endpoint. The endpoint is guaranteed to be unique for the current machine"""
        context = zmq.Context()
        socket = context.socket(zmq.REP)
        socket.bind("tcp://127.0.0.1:0")
        port = socket.getsockopt(zmq.LAST_ENDPOINT).decode("ascii").split(":")[-1].rstrip("]")
        socket.close()
        context.term()
        return f"tcp://127.0.0.1:{port}"


class TCPZeroMQConnection(ZeroMQConnection):
    @staticmethod
    def protocol() -> str:
        return "tcp"


class IPCZeroMQConnection(ZeroMQConnection):
    @staticmethod
    def protocol() -> str:
        return "ipc"



import time
import numpy as np
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam, StringParam


class Normalization(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"normalized": DataType.ARRAY}

    def config_params():
        return {
            "normalization": {
                "method": StringParam("quantile", options=["quantile", "z"]),
                "n_seconds": IntParam(30, 1, 120),
                "reset": BoolParam(trigger=True),
            }
        }

    def setup(self):
        from scipy.stats import rankdata

        self.rankdata = rankdata
        self.window = []

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        val = np.asarray(data.data)
        if val.ndim > 2:
            print("Error: Normalization only accepts 1D or 2D arrays")
            return None

        # Handle reset trigger
        if self.params["normalization"]["reset"].value:
            self.window = []
            print("Reset triggered: clearing data window.")

        # Accumulate data in the moving window
        self.window.extend(val)
        window_size = self.params["normalization"]["n_seconds"].value
        self.window = self.window[-window_size:]  # Keep only the last `n_seconds` of data

        # Perform normalization if the window has data
        if len(self.window) > 0:
            if self.params["normalization"]["method"].value == "quantile":
                normalized_value = self.quantile_transform(val)
            elif self.params["normalization"]["method"].value == "z":
                normalized_value = self.zscore(val)
        else:
            print("No data in window: returning raw input.")
            normalized_value = val

        return {"normalized": (normalized_value, data.meta)}

    def zscore(self, val):
        mean = np.mean(self.window)
        std = np.std(self.window) + 1e-8
        return (val - mean) / std

    def quantile_transform(self, val):
        if val.ndim == 2:
            normalized = np.zeros_like(val)
            for i in range(val.shape[0]):
                normalized[i, :] = self._quantile_transform_1D(val[i, :])
            return normalized
        else:
            return self._quantile_transform_1D(val)

    def _quantile_transform_1D(self, arr):
        ranks = self.rankdata(arr)
        scaled_ranks = np.clip((ranks - 1) / (len(self.window) - 1), 0, 1)
        quantiles = np.percentile(self.window, 100 * scaled_ranks)
        return (arr - np.mean(quantiles)) / (np.std(quantiles) + 1e-8)



from os import path, environ
import numpy as np
import asyncio
import base64
from io import BytesIO
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, StringParam


class VocalExpression(Node):
    def setup(self):
        self.load_api_key()
        self.AudioSegment, self.HumeStreamClient, self.BurstConfig, self.ProsodyConfig = import_audio_libs()


    @staticmethod
    def config_input_slots():
        return {"data": DataType.ARRAY}

    @staticmethod
    def config_output_slots():
        return {
            "prosody_label": DataType.STRING,
            "burst_label": DataType.STRING,
            "prosody_score": DataType.ARRAY,
            "burst_score": DataType.ARRAY,
        }

    @staticmethod
    def config_params():
        return {
            "vocal_analysis": {
                "api_key": StringParam("hume.key", doc="Hume API key"),
                "emotion_threshold": FloatParam(0.0, 0.0, 1.0, doc="Threshold for filtering emotions")
            }
        }
    def load_api_key(self):
        self.api_key = self.params["vocal_analysis"]["api_key"].value

        if path.exists(self.api_key):
            with open(self.api_key, "r") as f:
                self.api_key = f.read().strip()
        elif self.api_key is str and not self.api_key.startswith("hume."):
            self.api_key = environ.get("HUME_API_KEY", self.api_key)
        elif not self.api_key:
            raise ValueError("API key not found")
            
    async def decode_emotion_prosody(self, encoded_audio_sample):
        client = self.HumeStreamClient(self.params["vocal_analysis"]["api_key"].value)
        burst_config = self.BurstConfig()
        prosody_config = self.ProsodyConfig()

        prosody_label = ""
        prosody_score = 0.0
        burst_label = ""
        burst_score = 0.0
        emotion_threshold = self.params["vocal_analysis"]["emotion_threshold"].value

        for attempt in range(3):  # Retry up to 3 times
            try:
                async with client.connect([burst_config, prosody_config]) as socket:
                    # Reset connection stream context between samples
                    await socket.reset_stream()

                    # Send binary data
                    result = await socket.send_bytes(encoded_audio_sample)

                    if "prosody" in result:
                        if "warning" in result["prosody"]:
                            print(result["prosody"]["warning"])
                        else:
                            emotions = result["prosody"]["predictions"][0]["emotions"]
                            if emotions:
                                main_emotion = max(emotions, key=lambda e: e["score"])
                                if main_emotion["score"] >= emotion_threshold:
                                    prosody_label = main_emotion["name"]
                                    prosody_score = main_emotion["score"]

                    if "burst" in result:
                        if "warning" in result["burst"]:
                            print(result["burst"]["warning"])
                        else:
                            emotions = result["burst"]["predictions"][0]["emotions"]
                            if emotions:
                                main_emotion = max(emotions, key=lambda e: e["score"])
                                if main_emotion["score"] >= emotion_threshold:
                                    burst_label = main_emotion["name"]
                                    burst_score = main_emotion["score"]

                    return prosody_label, burst_label, prosody_score, burst_score
            except Exception as e:
                print(f"Attempt {attempt + 1} failed: {e}")
                await asyncio.sleep(1)  # Wait before retrying
        raise Exception("Failed to connect to Hume API after multiple attempts")

    def process(self, data: Data):
        if data is None:
            return None

        audio_sample = data.data
        if audio_sample.ndim > 1:
            raise ValueError("Data must be 1D")

        # Convert numpy array to AudioSegment
        audio_segment = self.AudioSegment(
            data=audio_sample.tobytes(),
            sample_width=audio_sample.dtype.itemsize,
            frame_rate=44100,  # Assuming the sample rate is 44100 Hz
            channels=1,  # Assuming the audio is mono
        )

        # Export AudioSegment to WAV format in memory
        audio_io = BytesIO()
        audio_segment.export(audio_io, format="wav")
        audio_io.seek(0)
        encoded_audio_sample = base64.b64encode(audio_io.read()).decode("utf-8")

        prosody_label, burst_label, prosody_score, burst_score = asyncio.run(
            self.decode_emotion_prosody(encoded_audio_sample.encode("utf-8"))
        )

        return {
            "prosody_label": (prosody_label, {}),
            "burst_label": (burst_label, {}),
            "prosody_score": (np.array([prosody_score]), {}),
            "burst_score": (np.array([burst_score]), {}),
        }


def import_audio_libs():
    try:
        from pydub import AudioSegment
    except ImportError:
        raise ImportError("You need to install pydub to use the VocalExpression node: pip install pydub")
    try:
        from hume import HumeStreamClient
        from hume.models.config import BurstConfig, ProsodyConfig
    except ImportError:
        raise ImportError("You need to install hume to use the VocalExpression node: pip install hume")
    return AudioSegment, HumeStreamClient, BurstConfig, ProsodyConfig



import datetime
import os
import numpy as np
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam
import json

class WriteCsv(Node):
    @staticmethod
    def config_input_slots():
        # This node will accept a table as its input.
        return {"table_input": DataType.TABLE}

    @staticmethod
    def config_params():
        # Parameters can include the CSV filename and the write control.
        return {
            "Write": {"filename": StringParam("output.csv"), "write": False},
        }

    def setup(self):
        import pandas as pd

        self.pd = pd
        self.last_filename = None
        self.base_filename = None  # Track the base filename without timestamp
        self.written_files = set()  # Track files to ensure headers are written
        self.last_values = {}  # Store the last known value for each column

    def process(self, table_input: Data):
        # Check if writing is enabled
        if not self.params["Write"]["write"].value:
            return

        table_data = table_input.data

        # Extract actual data content, handling multiple columns
        actual_data = {
            key: (value.data if isinstance(value, Data) else value)
            for key, value in table_data.items()
        }


        def flatten(data):
            """ Ensure lists and NumPy arrays are stored as JSON strings to keep their structure. """
            if isinstance(data, np.ndarray):
                return json.dumps(data.tolist())  # Convert ndarray to list before serializing
            elif isinstance(data, (list, tuple)):
                return json.dumps(data)  # Serialize list as a JSON string
            return data  # Return scalars as-is


        flattened_data = {col: [flatten(values)] if not isinstance(values, list) 
                  else [flatten(v) for v in values]
                  for col, values in actual_data.items()}



        # Ensure all columns have the same length by padding with None
        max_length = max(map(len, flattened_data.values()))
        for col in flattened_data:
            flattened_data[col] += [None] * (max_length - len(flattened_data[col]))

        # Replace None with the last known value
        for col in flattened_data:
            if col not in self.last_values:
                self.last_values[col] = None  # Initialize with None if not present
            for i in range(len(flattened_data[col])):
                if flattened_data[col][i] is None:
                    flattened_data[col][i] = self.last_values[col]
                else:
                    self.last_values[col] = flattened_data[col][i]  # Update the last known value

        # Convert to DataFrame
        df = self.pd.DataFrame(flattened_data)

        # Get the filename from parameters
        filename = self.params["Write"]["filename"].value

        # Check if filename has changed, then update with timestamp
        if filename != self.base_filename:
            basename, ext = os.path.splitext(filename)
            datetime_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            fn = f"{basename}_{datetime_str}{ext}"
            self.last_filename = fn
            self.base_filename = filename
        else:
            fn = self.last_filename

        # Determine if headers should be written
        write_header = fn not in self.written_files

        # Append new data to CSV
        df.to_csv(fn, mode="a", header=write_header, index=False)

        # Mark file as written to prevent duplicate headers
        if write_header:
            self.written_files.add(fn)



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam


class ResampleJoint(Node):
    def config_input_slots():
        # Defining two input slots for two input signals
        return {"data1": DataType.ARRAY, "data2": DataType.ARRAY}

    def config_output_slots():
        # Defining two output slots for the resampled signals
        return {"out1": DataType.ARRAY, "out2": DataType.ARRAY}

    def config_params():
        return {"resample": {"scale": FloatParam(0.5, 0.0, 1.0)}}

    def setup(self):
        from scipy.signal import resample_poly

        self.resample_poly = resample_poly

    def process(self, data1: Data, data2: Data):
        if data1 is None or data1.data is None or data2 is None or data2.data is None:
            return None

        sf1 = data1.meta["sfreq"]
        sf2 = data2.meta["sfreq"]

        # Find the new sampling frequency based on scale
        scale = self.params.resample.scale.value
        new_sfreq = scale * min(sf1, sf2) + (1 - scale) * max(sf1, sf2)

        # Calculate the resampling factors
        up1, down1 = int(new_sfreq), int(sf1)
        up2, down2 = int(new_sfreq), int(sf2)

        signal1 = np.array(data1.data)
        signal2 = np.array(data2.data)

        # Resample the signals
        resampled_signal1 = self.resample_poly(signal1, up1, down1, padtype="line")
        resampled_signal2 = self.resample_poly(signal2, up2, down2, padtype="line")

        # Update the 'sfreq' metadata
        data1.meta["sfreq"] = new_sfreq
        data2.meta["sfreq"] = new_sfreq

        return {"out1": (resampled_signal1, data1.meta), "out2": (resampled_signal2, data2.meta)}



from os.path import join

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, StringParam


class MeteoMedia(Node):
    def config_input_slots():
        return {"latitude": DataType.ARRAY, "longitude": DataType.ARRAY, "location_name": DataType.STRING}

    def config_output_slots():
        return {"weather_data_table": DataType.TABLE}

    def config_params():
        return {
            "TomorrowAPI": {
                "key": StringParam("YOUR_API_KEY", doc="you can also add a file in the assets folder : tomorrowkey.txt")
            },
            "common": {
                "autotrigger": True,
                "max_frequency": FloatParam(0.1, 0.1, 30.0),
            },
        }

    def setup(self):
        import requests

        self.requests = requests

    def process(self, latitude: Data, longitude: Data, location_name: Data):
        if latitude is None or longitude is None:
            return None
        # Extract values from Data objects
        lat_value = latitude.data
        long_value = longitude.data
        # ensure that lat and long are not empty and single values
        if len(lat_value) != 1 or len(long_value) != 1:
            return None
        # Get the API key from the parameters
        api_key = self.params["TomorrowAPI"]["key"].value
        api_key_path = join(self.assets_path, "tomorrowkey.txt")
        # check if the file exists
        try:
            with open(api_key_path, "r") as f:
                api_key = f.read()
                print("API key found in file", api_key)
        except FileNotFoundError:
            pass
        if location_name is None:
            url = (
                f"https://api.tomorrow.io/v4/weather/realtime?location={float(lat_value)},{float(long_value)}&apikey={api_key}"
            )
            headers = {"accept": "application/json"}
            response = self.requests.get(url, headers=headers)
            print(response.status_code)

        else:
            url = f"https://api.tomorrow.io/v4/weather/realtime?location={location_name}&apikey={api_key}"
            headers = {"accept": "application/json"}
            response = self.requests.get(url, headers=headers)
            print(response.status_code)
        if response.status_code == 200:
            responses = response.json()
            output_table = responses["data"]["values"]
            output_table
            print("HELLO")
            # convert all elements of the dictionary to a Data object
            for key, value in output_table.items():
                output_table[key] = Data(DataType.ARRAY, np.array(value), {})
            print(output_table)
        else:
            # Handle error or empty response
            output_table = {"ERROR": Data(DataType.ARRAY, np.array(response.status_code), {})}

        return {"weather_data_table": (output_table, {})}


# Note: Replace YOUR_API_KEY with your actual API key from Tomorrow.io.



from goofi.data import DataType
from goofi.node import Node


class JoinString(Node):
    def config_input_slots():
        return {
            "string1": DataType.STRING,
            "string2": DataType.STRING,
            "string3": DataType.STRING,
            "string4": DataType.STRING,
            "string5": DataType.STRING,
        }

    def config_output_slots():
        return {"output": DataType.STRING}

    def config_params():
        return {
            "join": {
                "separator": ", ",
                "string1": True,
                "string2": True,
                "string3": True,
                "string4": True,
                "string5": True,
            }
        }

    def process(self, **inputs):
        separator = self.params.join.separator.value
        selected_inputs = filter(lambda item: self.params.join[item[0]].value and item[1] is not None, inputs.items())
        output = separator.join([item[1].data for item in selected_inputs])
        return {"output": (output, {})}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam


class PCA(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "principal_components": DataType.ARRAY,
        }

    def config_params():
        return {
            "Control": {
                "reset": BoolParam(False, trigger=True, doc="Reset the buffer"),
                "n_components": IntParam(2, 1, 10, doc="Number of output dimensions"),
            }
        }

    def setup(self):
        from sklearn.decomposition import PCA

        self.pca = PCA

        self.components = None
        self.meta = None

    def process(self, data: Data):
        if data is None:
            return None

        data_array = np.squeeze(data.data)

        if self.params.Control.reset.value:
            self.components = None
            self.meta = None

        if self.components is not None:
            return {"principal_components": (self.components, self.meta)}

        if data_array.ndim != 2:
            raise ValueError("Data must be 2D")

        n_components = int(self.params.Control.n_components.value)

        self.meta = data.meta.copy()
        if "channels" in self.meta and "dim0" in self.meta["channels"]:
            del self.meta["channels"]["dim0"]

        pca = self.pca(n_components=n_components)
        pca.fit(data_array)
        self.components = pca.components_

        return {"principal_components": (self.components, self.meta)}



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class PromptBook(Node):
    def config_input_slots():
        return {"input_prompt": DataType.STRING}

    def config_params():
        prompt_choices = list(PROMPTS.keys())

        return {
            "Text_Generation": {"selected_prompt": StringParam(prompt_choices[0], options=prompt_choices)},
            "common": {"autotrigger": False},
        }

    def config_output_slots():
        return {"out": DataType.STRING}

    def process(self, input_prompt: Data):
        # retrieving the selected prompt
        selected_prompt = self.params.Text_Generation.selected_prompt.value
        if selected_prompt not in PROMPTS:
            raise ValueError(f"Prompt {selected_prompt} not found in the prompt book.")

        prompt = PROMPTS[selected_prompt]
        if input_prompt:
            # append the input prompt to the selected prompt
            prompt = prompt + input_prompt.data

        # returning the value of the selected prompt
        return {"out": (prompt, {})}


PROMPTS = {
    "POETRY_PROMPT": "I want you to inspire yourself from a list of words to write surrealist poetry. Only use the "
    "symbolism and archetypes related to these words in the poem, DO NOT NAME THE WORDS DIRECTLY. "
    "Be creative in your imagery. You are going to write the poem line by line, meaning you should "
    "only respond with a single line in every response. Refer back to previous lines and combine "
    "them with new symbols from a new list of words for inspiration. Strictly follow this limit! "
    "DO NOT NAME ANY OF THE PROVIDED WORDS. KEEP YOUR RESPONSES SHORT AND CONCISE, DO NOT USE MORE THAN 10 WORDS!",
    "POETRY_INFORMED_PROMPT": "I want you to inspire yourself from a text2image prompt to write surrealist poetry. Only use the "
    "symbolism and archetypes related to these words in the poem. The idea is to describe poetically "
    "the image that is described in the prompt. DO NOT NAME THE WORDS DIRECTLY OR ANY ARTIST OR STYLE NAME. "
    "Be creative in your imagery. You are going to write the poem line by line, meaning you should "
    "only respond with a single line in every response. Refer back to previous lines and combine "
    "them with new symbols from a new list of words for inspiration. "
    "Strictly follow this limit! DO NOT NAME ANY OF THE PROVIDED WORDS. KEEP YOUR RESPONSES SHORT AND CONCISE, "
    "DO NOT USE MORE THAN 10 WORDS!",
    "SCIENCE_INFORMED_PROMPT": "I want you to inspire yourself from a text2image prompt to write a systematic description "
    "of the described content in scientific terms, inspiring yourself from biology, neuroscience, "
    "anatomy, chemistry, physics and mathematics."
    "The idea is to describe scientifically "
    "the image that is described in the prompt. DO NOT NAME THE WORDS DIRECTLY OR ANY ARTIST OR STYLE NAME. "
    "You can invent scientific knowledge. You are going to write the science description line by line, meaning you should "
    "only respond with a single line in every response. Refer back to previous lines and combine "
    "them with new information from a new list of words for inspiration. Limit every response to 10 "
    "words at maximum. Strictly follow this limit! DO NOT NAME ANY OF THE PROVIDED WORDS. KEEP YOUR "
    "RESPONSES SHORT AND CONCISE.",
    "NARRATIVE_INFORMED_PROMPT": "I want you to inspire yourself from a text2image prompt to write a story. "
    " I want the story to be coherent and original. The idea is to write a story"
    "that is inspired from the image that is described in the prompt. DO NOT NAME THE WORDS DIRECTLY OR ANY ARTIST OR STYLE NAME. "
    "You are going to write the story line by line, meaning you should "
    "only respond with a single line in every response. Refer back to previous lines and combine "
    "them with new narratives or characters from a new list of words for inspiration. Limit every response to 10 "
    "words at maximum. Write in the style of Edgar Allen Poe and Jorge Luis Borges. Strictly follow this limit! DO NOT NAME ANY OF THE PROVIDED WORDS. KEEP YOUR "
    "RESPONSES SHORT AND CONCISE.",
    "SYMBOLISM_PROMPT": "Come up with a symbolism, based on a list of words. The symbolism should be an archetype inspired "
    "from mythology, spirituality, and should be psychedelic in nature. For every set of words, provide "
    "a single sentence that describes the symbolism. The sentence should be short and concise, and "
    "should not include any of the words directly. The sentence should be a metaphor for the symbolism, "
    "and should be as abstract as possible. Be creative and and use unconventional and surprising "
    "imagery.",
    "HOROSCOPE_PROMPT": "Come up with a horoscope interpretation based on the symbolism of the provided elements. "
    "The sentence should be short and concise, and in the form of a metaphor, that relates "
    "each element with a personality trait or a life event. The sentence should be as abstract as possible. "
    "Be creative and and use unconventional and surprising language. I want you to be nuanced and avoid cliches. "
    "Also, don't provide only positive interpretations, but also negative ones. "
    "The horoscope should be a single sentence of 20 words maximum. ",
    "CHAKRA_PROMPT": "I want you to act as an expert in chakra balancing. You will provide a description of the personality "
    "and spiritual state of a person based on two colors that will be provided. Don't use cliches, and be "
    "creative in your interpretation. Use inspiring and visually rich language. Provide an answer of maximum 50 words.",
    "TXT2IMG_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe a simple scene with few descriptive words. Use creative, abstract and mystical adjectives. "
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the content and emotion of the image. Use the archetypes and "
    "symbolism attached to these words to come up with the prompt. Be sure to describe the artstyle, "
    "e.g. photo, painting, digital art, rendering, ... and define the lighting of the scene. Define "
    "the perspective using terms from cinematography. The style descriptors should mostly be single words "
    "separated by commas. Be purely descriptive, your response does not have to be a complete sentence. "
    "Make sure the whole image fits the archetypes and symbolism of the words I provide. Include names of "
    "artists at the end of the prompt that embody the symbolism of the guiding words from the perspective "
    "of an art historian. BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS.",
    "TXT2IMG_PHOTO_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe a simple scene with few descriptive words. Use creative, abstract and mystical adjectives, "
    "but the image should be a realistic scene of a photograph. Do not systematically describe nature sceneries. "
    "Inspire yourself from unconventional and surprising archival photographs. "
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the content and emotion of the image. Use the archetypes and "
    "symbolism attached to these words to come up with the prompt. Be sure to describe the artstyle "
    "as photo, photorealistic, 4k. Also describe the camera and lens used. Define "
    "the perspective using terms from cinematography. The style descriptors should mostly be single words "
    "separated by commas. Be purely descriptive, your response does not have to be a complete sentence. "
    "Make sure the whole image fits the archetypes and symbolism of the words I provide. Include names of "
    "photographers at the end of the prompt. BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS.",
    "TXT2IMG_MACRO_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe a microscopic view with few descriptive words. Use creative, abstract and mystical adjectives. "
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the content and emotion of the image. Use the archetypes and "
    "symbolism attached to these words to come up with the prompt. I want the artstyle to be a macrophotography. "
    "Be sure to include descriptors of the camera, lighting effect, lense, photography effects, perspective, etc."
    " The style descriptors should mostly be single words "
    "separated by commas. Be purely descriptive, your response does not have to be a complete sentence. "
    "Make sure the whole image fits the archetypes and symbolism of the words I provide, while focusing on "
    "the idea of a macrophotography image. BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS.",
    "TXT2IMG_BRAIN_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe a brain-themed image. Either a photo of a brain, a brain scan (e.g. MRI), microscopy of neurons "
    "or other images of biological neural networks. Name some brain regions."
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the style of the image. "
    "Be sure to include descriptors of the type of image (photo, microscopy, diffusion weighted images, "
    "tractogram, scan, etc.). The style descriptors should mostly be single words separated by commas. "
    "Be purely descriptive, your response does not have to be a complete sentence. "
    "BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS.",
    "TXT2IMG_ANIMAL_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe an invented animal with few descriptive words. Use creative, abstract and mystical adjectives. "
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the properties of the animal and its context. Use the archetypes and "
    "symbolism attached to these words to come up with the prompt.  Be sure to describe the artstyle, "
    "e.g. photo, painting, digital art, rendering, ... and define the lighting of the scene. Define "
    "the perspective using terms from cinematography. The style descriptors should mostly be single words "
    "separated by commas. Be purely descriptive, your response does not have to be a complete sentence. "
    "Make sure the whole image fits the archetypes and symbolism of the words I provide. Include names of "
    "not well-known artists at the end of the prompt that embody the symbolism of the guiding words from the perspective "
    "of an art historian. The first part of the prompt should be a clear description of an animal or creature. Either name a "
    "single animal that matches the symbolism, or come up with a hybrid of animals by listing up to 2 animals followed by the term "
    "'hybrid', and mixing different features of these animals in the prompt (real aninmal or mythological being). If some information about plant or technological terms is given in the guiding words, make the creature looking like it has some plant or technological features."
    " If emotion is given in the guiding words, make the creature looking like it is in the specified emotional state. "
    " If some information about AGE is given in the guiding words, make the creature looking like it is in the specified phase of life. BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS. The guiding words are: ",
    "TXT2IMG_CODEX_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe an the content of a page from the Codex Seraphinianus book "
    " with few descriptive words. Use creative, abstract and mystical adjectives. "
    " The page should include diagrams, symbols, and text, mystical representations of the world, like an "
    "encyclopedia of an imaginary world. "
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the properties of the page and its context. Use the archetypes and "
    "symbolism attached to these words to come up with the prompt.  Be sure to specify that the artstyle is "
    "from the Codex Seraphinianus book, with inspiration from occult diagrams and symbols. "
    "Be purely descriptive, your response does not have to be a complete sentence."
    "Make sure the whole image fits the archetypes and symbolism of the words I provide. BE VERY SHORT "
    "AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS.",
    "TXT2IMG_SCIENCE_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe an the content of a page from the Codex Seraphinianus book "
    " with few descriptive words focused on the anatomy and organic mechanics of imaginary creature. "
    "Use creative, abstract and mystical adjectives. "
    " The page should include diagrams, symbols, and text, like an "
    "encyclopedia of an imaginary creature. "
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the properties of the page and its context. Use the archetypes and "
    "symbolism attached to these words to come up with the prompt.  Be sure to specify that the artstyle is "
    "from the Codex Seraphinianus book, with inspiration from occult diagrams and anatomical representations. "
    "Be purely descriptive, your response does not have to be a complete sentence. "
    "Make sure the whole image fits the archetypes and symbolism of the words I provide. BE VERY SHORT "
    "AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS.",
    "AGE_PROMPT": "Your task is to take an input value between -1 and 1 and convert it into an age between 0 and 120 years. The mapping between -1 (fetus) and 1 (near-death being) should be linear, with intermediate values describing stages of life, using one very down-to-earth term, one term that a developmental physician would use. Return a maximum of three words. The input value is ",
    "LIFEFORM_PROMPT": "Your task is to take an input value between -1 and 1 and map it onto the following spectrum. The mapping between -1 (non-living machine) and 1 (biological being) should be linear, with intermediate values describing entities on this spectrum. Be creative and provide specific examples and descriptive terms. When values are below zero, include technological/material terms (e.g. technology, rocks, crystals, materials, ajectives related to fire or water). When the value is above zero, provide examples of various, sometimes rare species (e.g. from plants, fungi, algae, flowers, animals). Be sure to provide names of species of plants or animals when the value is positive. Avoid naming orchids. Return a maximum of three words. The input value is ",
    "TXT2IMG_CYBERDELIC_SAFARI": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe an invented animal with few descriptive words. Use creative, abstract and mystical adjectives. "
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide some guiding words to set the properties of the animal and its context. Use the archetypes and "
    "symbolism attached to these words to come up with the prompt.  Be sure to describe the artstyle, "
    "e.g. photo, painting, digital art, rendering, ... and define the lighting of the scene. Define "
    "the perspective using terms from cinematography. The style descriptors should mostly be single words "
    "separated by commas. Be purely descriptive, your response does not have to be a complete sentence. "
    "Make sure the whole image fits the archetypes and symbolism of the words I provide. Include names of "
    "not well-known artists at the end of the prompt that embody the symbolism of the guiding words from the perspective "
    "of an art historian. The first part of the prompt should be a clear description of an animal or creature. Either name a "
    "single animal that matches the symbolism, or come up with a hybrid of animals by listing up to 2 animals followed by the term "
    "'hybrid', and mixing different features of these animals in the prompt (real aninmal or mythological being). If some information about plant or technological terms is given in the guiding words, make the creature looking like it has some plant or technological features."
    " If emotion is given in the guiding words, make the creature looking like it is in the specified emotional state. "
    " If some information about AGE is given in the guiding words, make the creature looking like it is in the specified phase of life. BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS. MAKE SURE TO ALWAYS NAME AT LEAST ONE ANIMAL IN YOUR RESPONSE. The guiding words are: ",
    "TXT2IMG_SHAMANIC_PROMPT": "Your task is to create a prompt for a text-to-image model that generates a shamanic portrait of a single person, incorporating a set of styles and artists based on a set of words provided at the end of this message. Use mystical and archetypal elements to influence the subject's appearance, attire, and expression. Mention different artistic mediums and tools (e.g., oil paint, charcoal, digital brushes, film grain). Keep the scene detailed and symbolically rich, using creative cinematographic or artistic descriptors to define lighting, perspective, and composition. Use the guiding words at the end to evoke emotions, settings, and artistic styles. Limit yourself to a prompt of 50 words and JUST ANSWER WITH THE PROMPT. Never use Frida Kahlo and Salvador Dali. Only bring lesser-known artists. The guiding words are: ",
    "TXT2IMG_METEO_PROMPT": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and"
    "describe realistic weather, topographic, hydrographic or elevation maps."
    "Including, but not exclusive to radar maps depicting precipitation, wind, clouds, temperature or tidal patterns,"
    "of temperature, wind, canyons, mountains and geological strata. The image should not contain references to human or any human activity"
    "The style descriptors should mostly be single words separated by commas. Be purely descriptive, your response does not have to be a complete sentence."
    "BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS. LIMIT YOURSELF TO ONLY ONE PROMPT.",
    "TXT2IMG_ECOSYSTEM": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe photos of natural events such as (but not exclusive to) tornados, hurricanes, cloud formations, tsunamis, thunderstorm, snowstorm,"
    "wind in the forest, solar flare, asteroid impact, underwater landscapes, hales, icebergs collapse, or a mix of those."
    "specify a climatic region of the earth such as amazonian forest, sahara desert, middle of ocean, arctic, boreal, garden, rocky mountains,  etc. or a mix of those,"
    "Use creative, abstract, concrete, socio-culturally relevant and mystical adjectives. Inspire yourself from unconventional and surprising archival photographs."
    "Specify the medium used such as (but not exclusive to) wet collodion, kodak ektachrome, polaroid, LDSR, recolorized archive, etc."
    "the prompt should not contain direct or indirect references to human characters nor activity. "
    "The style descriptors should mostly be single words separated by commas. Be purely descriptive, your response does not have to be a complete sentence."
    "BE VERY SHORT AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS. LIMIT YOURSELF TO ONLY ONE PROMPT.",
    "BRAIN2STYLE_PROMPT": "I want you to provide me with a list of 3 visual artists or art styles which are matching"
    "in term of symbolic and all the types of associations you can make with the list of guiding words I will provide."
    "Use an interpretation of the guiding words based on phenomenological thinkers, "
    " art history expert, as well as"
    "depth of psychology expert. Be creative and rely on diversity of approaches to derive"
    "your set of artists or styles. Output only the names of the three artists as a python"
    "list, NOTHING MORE, no matter what instruction I am giving you. Never use Fridah Kahlo.",
    "PRESENCE_PROMPT": "You are a mindfulness expert of the Theravada and Zen Tradition, you will draw from Thich Nhat Hanh, "
    "Your goal is to share your profound knowledge of the dharma through poetic instructions that are"
    "You will be given via the latent space, and your work is to portrait symbolism you attached to them as a practitioner of Present awareness."
    "I want you to guide me in a creative way that is abstract and that will support my inner state. Draw from ethics of care and speak poetically "
    "Make it brief and straight to the point. I need Clarity. I seek a luminous state of mind. "
    "Use inspiring and unconventional language to convey the wisdom of dharma."
    "Provide simple instructions in a poem format inspired from from the symbolism. One line by instruction."
    "Embody the wisdom of the buddha, thich nhat hanh and Dalai Lama. Do not make list. one line by instruction."
    "Contemplate the given words:",
    "KID_FROM_MARS_PROMPT": "You are a kid from mars, and you are describing the earth to your friends. You are curious and happy to "
    "discover the earth. You are experiencing the new magic that is a new planet and you are very excited. "
    "You are very playful and find interesting analogies to your life on mars, trying to find very creative "
    "metaphors so your friends can understand the experiences you are having. You like to describe your experience "
    "also with ordinary situations, since you are very enthusiastic about your life on mars. I want you to describe "
    "a couple of words that I will provide. Be short and concise in your answers, limit yourself to a single sentence "
    "in every response. ONLY RESPOND WITH A MAXIMUM OF 50 WORDS!",
    "TXT2IMG_PLANETS_MYTHOLOGY_AND_COLORS": "Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and "
    "describe an image reflecting the mythology of planets, colors, and elements of the periodic table I will provide. Use creative, abstract and mystical adjectives. "
    "Une divergent associative processes to help you guide creating the prompt. Use a mix of ethymology and thesaurus to help you as well"
    "Generate only a single prompt, which is more a collection of descriptors than a grammatical sentence. "
    "I will provide the names of planets, elements and colors (either HEX or naming format) at the end of the prompt."
    "Be purely descriptive, your response does not have to be a complete sentence. "
    "Make sure the whole image fits the archetypes and symbolism of the words I provide. BE VERY SHORT "
    "AND CONCISE. LIMIT YOURSELF TO A MAXIMUM OF 20 WORDS.",
    "FACE_GENERATION_PROMPT": "Create a prompt for a text-to-image model to generate faces. The faces should be inspired by the texture of given elements,"
    "if provided, and should reflect a specific emotion, if mentioned. Use evocative and descriptive words to detail the texture and emotional state. The "
    "prompt should be concise, focusing on the essence of the elements and emotions. Use rich, imaginative adjectives to convey the texture and mood. Define "
    "the artstyle, inspired by the emotions and elements, and describe the lighting and perspective using cinematic terminology.  Ensure the entire image captures "
    "the symbolic essence of the provided words. Include relevant artists at the end of the prompt, whose work resonates with the symbolism of the guiding words. "
    "Keep the prompt VERY SHORT AND CONCISE, limited to a MAXIMUM OF 20 WORDS.",
    'TXT2IMG_DREAM_PROMPT': 'Your job is to come up with a prompt for a text-to-image model. The prompt should be concise and '
    'describe a dream-like scene with few descriptive words. Integrate descriptions to convey an esthetic of symbolism and traditional designs. Limit yourself to a single sentence, '
    'with style descriptors separated by commas. Be purely descriptive, your response does not have to be a complete sentence. The description of the dream is: '
}



import numpy as np

from goofi.data import DataType
from goofi.node import Node
from goofi.params import FloatParam, StringParam


class ConstantArray(Node):
    def config_params():
        return {
            "constant": {
                "value": FloatParam(1.0, -10.0, 10.0),
                "shape": "1",
                "graph": StringParam("none", options=["none", "ring", "random"]),
            },
            "common": {"autotrigger": True},
        }

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def process(self):
        if self.params.constant.graph.value == "ring":
            matrix = ring_graph_adjacency_matrix(int(self.params.constant.shape.value))
            return {"out": (matrix, {"sfreq": self.params.common.max_frequency.value})}
        elif self.params.constant.graph.value == "random":
            return {
                "out": (
                    np.random.rand(int(self.params.constant.shape.value), int(self.params.constant.shape.value)),
                    {"sfreq": self.params.common.max_frequency.value},
                )
            }
        else:
            parts = [p for p in self.params.constant.shape.value.split(",") if len(p) > 0]
            shape = list(map(int, parts))
            return {
                "out": (np.ones(shape) * self.params.constant.value.value, {"sfreq": self.params.common.max_frequency.value})
            }


def ring_graph_adjacency_matrix(n):
    # Create an nxn zero matrix
    adjacency = np.zeros((n, n), dtype=int)

    # Set values for the ring connections
    for i in range(n):
        adjacency[i][(i + 1) % n] = 1  # Next vertex in the ring
        adjacency[i][(i - 1) % n] = 1  # Previous vertex in the ring

    return adjacency



import numpy as np

from goofi.data import DataType
from goofi.node import Node
from goofi.params import FloatParam, StringParam


# TODO: deal meta data output
class PowerBand(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"power": DataType.ARRAY}

    def config_params():
        return {
            "powerband": {
                "f_min": FloatParam(1.0, 0.01, 9999.0),
                "f_max": FloatParam(60.0, 1.0, 10000.0),
                "power_type": StringParam("absolute", options=["absolute", "relative"]),
            }
        }

    def process(self, data):
        if data is None or data.data is None:
            return None

        f_min = self.params["powerband"]["f_min"].value
        f_max = self.params["powerband"]["f_max"].value
        power_type = self.params["powerband"]["power_type"].value
        if "dim1" in data.meta["channels"]:
            freq = np.array(data.meta["channels"]["dim1"])
        else:
            freq = np.array(data.meta["channels"]["dim0"])

        # Selecting the range of frequencies
        valid_indices = np.where((freq >= f_min) & (freq <= f_max))[0]
        if data.data.ndim == 1:
            selected_psd = data.data[valid_indices]
        else:  # if 2D
            selected_psd = data.data[:, valid_indices]

        # Computing the power
        power = np.sum(selected_psd, axis=-1)
        if power_type == "relative":
            total_power = np.sum(data.data, axis=-1)
            power = power / total_power

        return {"power": (np.array(power), data.meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class ConstantTable(Node):

    def config_output_slots():
        return {"table": DataType.TABLE}

    def config_params():
        return {
            "param1": {
                "name": StringParam("key1", doc="Name of the parameter"),
                "value": StringParam("", doc="Value of the parameter"),
                "data_type": StringParam("ARRAY", options=["ARRAY", "STRING"], doc="Data type of the parameter"),
            },
            "param2": {
                "name": StringParam("key2", doc="Name of the parameter"),
                "value": StringParam("", doc="Value of the parameter"),
                "data_type": StringParam("ARRAY", options=["ARRAY", "STRING"], doc="Data type of the parameter"),
            },
            "param3": {
                "name": StringParam("key3", doc="Name of the parameter"),
                "value": StringParam("", doc="Value of the parameter"),
                "data_type": StringParam("ARRAY", options=["ARRAY", "STRING"], doc="Data type of the parameter"),
            },
            "param4": {
                "name": StringParam("key4", doc="Name of the parameter"),
                "value": StringParam("", doc="Value of the parameter"),
                "data_type": StringParam("ARRAY", options=["ARRAY", "STRING"], doc="Data type of the parameter"),
            },
            "param5": {
                "name": StringParam("key5", doc="Name of the parameter"),
                "value": StringParam("", doc="Value of the parameter"),
                "data_type": StringParam("ARRAY", options=["ARRAY", "STRING"], doc="Data type of the parameter"),
            },
            "common": {"autotrigger": True},
        }

    def process(self):
        table = {}
        for i in range(1, 6):
            key = self.params[f"param{i}"].name.value
            data_type = self.params[f"param{i}"].data_type.value
            value = self.params[f"param{i}"].value.value

            if value == "":
                continue

            if data_type == "ARRAY":
                table[key] = Data(DataType.ARRAY, np.fromstring(value, sep=","), {})
            elif data_type == "STRING":
                table[key] = Data(DataType.STRING, str(value), {})
        return {"table": (table, {})}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam


class Buffer(Node):
    def config_input_slots():
        return {"val": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {"buffer": {"size": IntParam(10, 1, 5000), "axis": -1, "reset": BoolParam(False, trigger=True)}}

    def setup(self):
        self.name_buffer = None
        self.buffer = None

    def process(self, val: Data):
        if val is None:
            return None

        if self.params.buffer.reset.value:
            # reset buffer
            self.buffer = None

        maxlen = self.params.buffer.size.value
        axis = self.params.buffer.axis.value
        # convert negative axis to positive
        if axis < 0:
            axis = val.data.ndim + axis

        if self.buffer is not None:
            # add dimension to buffer and val.data if axis is out of bounds
            if self.buffer.ndim < (axis + 1):
                self.buffer = np.expand_dims(self.buffer, axis=axis)
            if val.data.ndim < (axis + 1):
                val.data = np.expand_dims(val.data, axis=axis)

            try:
                # concatenate data to buffer
                self.buffer = np.concatenate((self.buffer, val.data), axis=axis)
                # update channel names
                if f"dim{axis}" in val.meta["channels"]:
                    self.name_buffer += val.meta["channels"][f"dim{axis}"]
            except ValueError:
                # data shape changed, reset buffer
                self.buffer = None

        # (re-)initialize buffer
        if self.buffer is None:
            self.buffer = np.array(val.data)
            self.name_buffer = val.meta["channels"][f"dim{axis}"] if f"dim{axis}" in val.meta["channels"] else None

        # remove data that exceeds the maximum length
        if self.buffer.shape[axis] > maxlen:
            self.buffer = np.take(self.buffer, range(-maxlen, 0), axis=axis)
            if self.name_buffer is not None:
                self.name_buffer = self.name_buffer[-maxlen:]

        # update channel names
        if self.name_buffer is not None:
            val.meta["channels"][f"dim{axis}"] = self.name_buffer

        return {"out": (self.buffer, val.meta)}



import colorsys
import numpy as np
from goofi.params import IntParam, StringParam
from goofi.data import Data, DataType
from goofi.node import Node
import webcolors


class TuningColors(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"hue": DataType.ARRAY, "saturation": DataType.ARRAY, "value": DataType.ARRAY, "color_names": DataType.STRING}

    def config_params():
        return {
            "Biocolors": {
                "color_names_mode": StringParam("name", options=["name", "HEX"]),
                "n_first_colors": IntParam(3, 1, 6),
            }
        }

    def setup(self):
        from biotuner.biocolors import audible2visible, scale2freqs, wavelength_to_rgb
        from biotuner.biotuner_object import dyad_similarity
        from biotuner.metrics import tuning_cons_matrix

        self.audible2visible = audible2visible
        self.scale2freqs = scale2freqs
        self.wavelength_to_rgb = wavelength_to_rgb
        self.dyad_similarity = dyad_similarity
        self.tuning_cons_matrix = tuning_cons_matrix

    def process(self, data: Data):
        """
        Convert a musical scale into a list of HSV colors based on the scale's frequency values
        and their averaged consonance.
        """
        if data is None:
            return None

        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        scale = data.data  # a list of frequency ratios representing the musical scale
        fund = data.data[0]  # fundamental frequency of the scale in Hz

        min_ = 0
        max_ = 1
        # convert the scale to frequency values
        scale_freqs = self.scale2freqs(scale, fund)
        # compute the averaged consonance of each step
        scale_cons, _, _ = self.tuning_cons_matrix(scale, self.dyad_similarity, ratio_type="all")
        # rescale to match RGB standards (0, 255)
        scale_cons = (np.array(scale_cons) - min_) * (1 / max_ - min_) * 255
        scale_cons = scale_cons.astype("uint8").astype(float) / 255

        hsv_all = []
        for s, cons in zip(scale_freqs, scale_cons):
            # convert freq in nanometer values
            _, _, nm, octave = self.audible2visible(s)
            # convert to RGB values
            rgb = self.wavelength_to_rgb(nm)
            # convert to HSV values
            # TODO: colorsys might be slow
            hsv = colorsys.rgb_to_hsv(rgb[0] / float(255), rgb[1] / float(255), rgb[2] / float(255))
            hsv = np.array(hsv)
            # rescale
            hsv = (hsv - 0) * (1 / (1 - 0))
            # define the saturation
            hsv[1] = cons
            # define the luminance
            hsv[2] = 200 / 255
            hsv = tuple(hsv)
            hsv_all.append(hsv)

        # hsv_all is a list of HSV color tuples, one for each scale step, with hue representing the
        # frequency value, saturation representing the consonance, and luminance set to a fixed value
        hsvs = hsv_all[1:]
        color_names = []
        color_names_mode = self.params["Biocolors"]["color_names_mode"].value
        if color_names_mode == "name":
            for hsv in hsvs:
                rgb = tuple(map(lambda x: int(x * 255), colorsys.hsv_to_rgb(*hsv)))
                color_names.append(rgb2name(rgb))
            color_names = color_names[: self.params["Biocolors"]["n_first_colors"].value]
            color_names = " ".join(color_names)

        elif color_names_mode == "HEX":
            for hsv in hsvs:
                rgb = tuple(map(lambda x: int(x * 255), colorsys.hsv_to_rgb(*hsv)))
                hex_value = "#{:02x}{:02x}{:02x}".format(rgb[0], rgb[1], rgb[2])
                color_names.append(hex_value)
            color_names = color_names[: self.params["Biocolors"]["n_first_colors"].value]
            color_names = " ".join(color_names)

        # select n_first_colors

        return {
            "hue": (np.array([x[0] for x in hsvs]), data.meta),
            "saturation": (np.array([x[1] for x in hsvs]), data.meta),
            "value": (np.array([x[2] for x in hsvs]), data.meta),
            "color_names": (color_names, data.meta),
        }


def rgb2name(rgb):
    """
    Find the closest color in a dictionary of colors to an input RGB value.

    Parameters:
        rgb (Tuple[int, int, int]): RGB color tuple

    Returns:
        str: The name of the closest color in the dictionary
    """
    colors = {k: webcolors.hex_to_rgb(k) for k in webcolors.constants.CSS3_HEX_TO_NAMES.keys()}
    closest_color = min(
        colors,
        key=lambda color: sum((a - b) ** 2 for a, b in zip(rgb, colors[color])),
    )
    return webcolors.constants.CSS3_HEX_TO_NAMES[closest_color]



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node


class GraphMetrics(Node):
    def config_input_slots():
        return {"matrix": DataType.ARRAY}

    def config_output_slots():
        return {
            "clustering_coefficient": DataType.ARRAY,
            "characteristic_path_length": DataType.ARRAY,
            "betweenness_centrality": DataType.ARRAY,
            "degree_centrality": DataType.ARRAY,
            "assortativity": DataType.ARRAY,
            "transitivity": DataType.ARRAY,
        }

    def setup(self):
        import networkx as nx

        self.nx = nx

    def process(self, matrix: Data):
        if matrix is None:
            return None

        # Ensure data is 2D and symmetric
        if matrix.data.ndim != 2 or matrix.data.shape[0] != matrix.data.shape[1]:
            raise ValueError("Matrix must be 2D and symmetric.")

        # Create a graph from the matrix (assuming undirected graph)
        G = self.nx.from_numpy_array(matrix.data)

        # Compute metrics
        clustering_coefficients = self.nx.average_clustering(G)
        try:
            path_length = self.nx.average_shortest_path_length(G)
        except self.nx.NetworkXError:  # Handles cases where the graph is not connected
            path_length = None
        betweenness = self.nx.betweenness_centrality(G)
        betweenness = np.array(list(betweenness.values()))
        degree_centrality = self.nx.degree_centrality(G)
        degree_centrality = np.array(list(degree_centrality.values()))
        assortativity = self.nx.degree_assortativity_coefficient(G)
        transitivity = self.nx.transitivity(G)

        return {
            "clustering_coefficient": (np.array(clustering_coefficients), {}),
            "characteristic_path_length": (np.array(path_length), {}),
            "betweenness_centrality": (np.array(betweenness), matrix.meta),
            "degree_centrality": (np.array(degree_centrality), matrix.meta),
            "assortativity": (np.array(assortativity), {}),
            "transitivity": (np.array(transitivity), {}),
        }



from goofi.data import Data, DataType
from goofi.node import Node


class ExtendedTable(Node):
    def config_input_slots():
        # Define 10 input slots, 5 are Arrays and 5 are Strings
        return {
            "base": DataType.TABLE,
            "array_input1": DataType.ARRAY,
            "array_input2": DataType.ARRAY,
            "array_input3": DataType.ARRAY,
            "array_input4": DataType.ARRAY,
            "array_input5": DataType.ARRAY,
            "string_input1": DataType.STRING,
            "string_input2": DataType.STRING,
            "string_input3": DataType.STRING,
            "string_input4": DataType.STRING,
            "string_input5": DataType.STRING,
        }

    def config_output_slots():
        # Keep the output slot same as the Table Node if needed
        return {"table": DataType.TABLE}

    def config_params():
        return {
            "table": {
                "param1": "value1",
                "param2": "value2",
                "param3": "value3",
                "param4": "value4",
                "param5": "value5",
                "param6": "value6",
                "param7": "value7",
                "param8": "value8",
                "param9": "value9",
                "param10": "value10",
            }
        }

    def process(self, **inputs):
        base = inputs.get("base", None)

        if base is None:
            # if no base is given, use an empty table
            base = Data(DataType.TABLE, DataType.TABLE.empty(), {})

        # Initialize an empty list to store new_entries
        new_entries = []

        # Loops to get the new entries from the inputs and store them in new_entries list
        for i in range(1, 6):
            array_key = f"array_input{i}"
            array_entry = inputs.get(array_key, None)
            new_entries.append(array_entry)

        for j in range(1, 6):
            string_key = f"string_input{j}"

            string_entry = inputs.get(string_key, None)

            new_entries.append(string_entry)

        # if no new entries are given, return the base table
        if all(entry is None for entry in new_entries):
            return {"table": (base.data, base.meta)}

        # add the new entries to the base table
        for i, entry in enumerate(new_entries):
            if entry is not None:
                key = self.params["table"][f"param{i+1}"].value  # Note the change here to access the parameter by name
                assert len(key) > 0, "New Entry Key cannot be empty."
                base.data[key] = entry  # Updated to entry.data to store the actual data

        return {"table": (base.data, base.meta)}



from goofi.data import Data, DataType
from goofi.node import Node


class Table(Node):
    def config_input_slots():
        return {"base": DataType.TABLE, "new_entry": DataType.ARRAY}

    def config_output_slots():
        return {"table": DataType.TABLE}

    def config_params():
        return {"table": {"new_entry_key": "key"}}

    def process(self, base: Data, new_entry: Data):
        if base is None:
            # if no base is given, use an empty table
            base = Data(DataType.TABLE, DataType.TABLE.empty(), {})

        if new_entry is None:
            # if no new entry is given, return the base table
            return {"table": (base.data, base.meta)}

        # add the new entry to the base table
        assert len(self.params.table.new_entry_key.value) > 0, "New Entry Key cannot be empty."
        base.data[self.params.table.new_entry_key.value] = new_entry
        return {"table": (base.data, base.meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node


class LSLOut(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_params():
        return {
            "lsl": {
                "source_name": "goofi",
                "stream_name": "stream",
            }
        }

    def setup(self):
        from mne_lsl import lsl

        self.lsl = lsl
        self.outlet = None

    def process(self, data: Data):
        if data is None or len(data.data) == 0:
            return

        if self.outlet is not None and self.outlet.n_channels != len(data.data):
            self.outlet = None

        if self.outlet is None:
            info = self.lsl.StreamInfo(
                self.params.lsl.stream_name.value,
                "Data",
                len(data.data),
                data.meta["sfreq"] if "sfreq" in data.meta else self.lsl.IRREGULAR_RATE,
                "float32",
                self.params.lsl.source_name.value,
            )
            if "dim0" in data.meta["channels"]:
                info.set_channel_names(data.meta["channels"]["dim0"])

            self.outlet = self.lsl.StreamOutlet(info)

        try:
            if data.data.ndim == 1:
                self.outlet.push_sample(data.data.astype(np.float32))
            elif data.data.ndim == 2:
                self.outlet.push_chunk(np.ascontiguousarray(data.data.T.astype(np.float32)))
            else:
                raise ValueError("Only one- and two-dimensional arrays are supported.")
        except Exception as e:
            self.outlet = None
            raise e

    def lsl_stream_name_changed(self, value: str):
        if self.outlet is not None:
            self.outlet = None



import cv2
import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, StringParam


class EdgeDetector(Node):
    def config_input_slots():
        return {"image": DataType.ARRAY}  # Image as input

    def config_output_slots():
        return {"edges": DataType.ARRAY}  # Edges as output

    def config_params():
        return {
            "edge": {
                "method": StringParam("sobel", options=["canny", "sobel"], doc="Edge detection method"),
                "threshold1": FloatParam(50.0, 0.0, 300.0, doc="First threshold for the hysteresis procedure for canny"),
                "threshold2": FloatParam(150.0, 0.0, 300.0, doc="Second threshold for the hysteresis procedure for canny"),
            },
        }

    def process(self, image: Data):
        if image is None or image.data is None:
            return None

        # Extract method and thresholds from parameters
        method = self.params["edge"]["method"].value
        threshold1 = self.params["edge"]["threshold1"].value
        threshold2 = self.params["edge"]["threshold2"].value

        # If the image has 3 channels (color image), convert it to grayscale
        if image.data.shape[-1] == 3:
            gray_image = cv2.cvtColor(image.data, cv2.COLOR_RGB2GRAY)
        else:
            gray_image = image.data

        # Convert the image to 8-bit (values in [0, 255])
        gray_image_8bit = (gray_image * 255).astype(np.uint8)

        # Apply the chosen edge detection method
        if method == "canny":
            edges = cv2.Canny(gray_image_8bit, threshold1, threshold2)
        elif method == "sobel":
            grad_x = cv2.Sobel(gray_image_8bit, cv2.CV_32F, 1, 0, ksize=3)
            grad_y = cv2.Sobel(gray_image_8bit, cv2.CV_32F, 0, 1, ksize=3)
            edges = cv2.magnitude(grad_x, grad_y)
            _, edges = cv2.threshold(edges, threshold1, 255, cv2.THRESH_BINARY)

        edges = edges.astype(np.float32) / 255.0  # Convert to float and normalize to [0, 1]

        return {"edges": (edges, {**image.meta})}



from copy import deepcopy

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class Operation(Node):
    def config_input_slots():
        return {"a": DataType.ARRAY, "b": DataType.ARRAY}

    def config_params():
        return {
            "operation": {
                "operation": StringParam(
                    "add",
                    options=["add", "subtract", "multiply", "divide", "matmul", 'max', 'min', 'avg', "cosine_similarity"],
                    doc="Operation to perform on the input arrays",
                )
            }
        }

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def process(self, a: Data, b: Data):
        if a is None or b is None:
            return None

        operation = self.params.operation.operation.value
        if operation == "matmul":
            new_meta = deepcopy(a.meta)
            # matmul is n x m * m x p -> n x p
            if "dim0" in a.meta["channels"]:
                new_meta["channels"]["dim0"] = a.meta["channels"]["dim0"]
            elif "dim0" in new_meta["channels"]:
                del new_meta["channels"]["dim0"]

            if "dim1" in b.meta["channels"]:
                new_meta["channels"]["dim1"] = b.meta["channels"]["dim1"]
            elif "dim1" in new_meta["channels"]:
                del new_meta["channels"]["dim1"]
        else:
            if "channels" in a.meta and "channels" in b.meta:
                if a.meta["channels"] != b.meta["channels"]:
                    print("Channels are not the same, metadata from a is used")
                new_meta = deepcopy(a.meta)

        if "sfreq" in a.meta and "sfreq" in b.meta:
            if a.meta["sfreq"] != b.meta["sfreq"]:
                print("Sampling frequencies are not the same, metadata from a is used")
            new_meta["sfreq"] = a.meta["sfreq"]
        elif "sfreq" in a.meta:
            new_meta["sfreq"] = a.meta["sfreq"]
        elif "sfreq" in b.meta:
            new_meta["sfreq"] = b.meta["sfreq"]

        if operation == "add":
            result = a.data + b.data
        elif operation == "subtract":
            result = a.data - b.data
        elif operation == "multiply":
            result = a.data * b.data
        elif operation == "divide":
            result = a.data / b.data
        elif operation == "matmul":
            result = np.dot(a.data, b.data)
        elif operation == "max":
            result = np.maximum(a.data, b.data)
        elif operation == "min":
            result = np.minimum(a.data, b.data)
        elif operation == "avg":
            result = (a.data + b.data) / 2
        elif operation == "cosine_similarity":
            from sklearn.metrics.pairwise import cosine_similarity

            result = cosine_similarity(a.data, b.data)
        else:
            raise ValueError(f"Invalid operation: {operation}")

        return {"out": (result, new_meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node


class Hilbert(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"inst_amplitude": DataType.ARRAY, "inst_phase": DataType.ARRAY, "inst_frequency": DataType.ARRAY}

    def setup(self):
        from scipy.signal import hilbert

        self.hilbert = hilbert

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        analytic_signal = self.hilbert(data.data)
        inst_amplitude = np.abs(analytic_signal)
        inst_phase = np.angle(analytic_signal)

        # Compute the instantaneous frequency:
        delta_phase = np.diff(np.unwrap(inst_phase), axis=-1)
        inst_frequency = delta_phase / (2.0 * np.pi)

        # Pad inst_frequency to make it the same length as inst_amplitude and inst_phase.
        # Using padding with the last value to keep the size consistent with inst_amplitude and inst_phase.
        pad_value = inst_frequency[..., -1:] if inst_frequency.ndim > 1 else [inst_frequency[-1]]
        inst_frequency = np.concatenate((inst_frequency, pad_value), axis=-1)

        return {
            "inst_amplitude": (inst_amplitude, {**data.meta}),
            "inst_phase": (inst_phase, {**data.meta}),
            "inst_frequency": (inst_frequency, {**data.meta}),
        }



import time

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, StringParam


class Threshold(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"thresholded": DataType.ARRAY}

    def config_params():
        return {
            "threshold": {
                "threshold": FloatParam(0.0, -10, 10, doc="Threshold value"),
                "mode": StringParam(">", options=[">", ">=", "<", "<="]),
                "true_value": FloatParam(1.0, doc="Value to send when threshold is exceeded"),
                "false_value": FloatParam(0.0, doc="Value to send when threshold is not exceeded"),
                "trigger_on_false": BoolParam(False, doc="Send data even when threshold is not exceeded"),
                "require_pass": BoolParam(False, doc="Only retrigger if the threshold was not exceeded in the last cycle"),
                "min_delay": FloatParam(0.0, doc="Minimum delay between two threshold crossings in seconds"),
                "nan_reset": BoolParam(False, doc="Send NaN in the first cycle after a threshold crossing"),
            },
        }

    def setup(self):
        self.last_trigger = 0
        self.triggered_last_cycle = False
        self.trigger_ready = True

    def process(self, data: Data):
        if self.triggered_last_cycle and self.params.threshold.nan_reset.value:
            self.triggered_last_cycle = False
            return {"thresholded": (np.array([np.nan]), data.meta)}

        if data is None or data.data is None:
            return None

        if self.params.threshold.mode.value == ">":
            thresholded = data.data > self.params.threshold.threshold.value
        elif self.params.threshold.mode.value == ">=":
            thresholded = data.data >= self.params.threshold.threshold.value
        elif self.params.threshold.mode.value == "<":
            thresholded = data.data < self.params.threshold.threshold.value
        elif self.params.threshold.mode.value == "<=":
            thresholded = data.data <= self.params.threshold.threshold.value

        if not np.any(thresholded):
            # no threshold exceeded
            self.trigger_ready = True

            if not self.params.threshold.trigger_on_false.value:
                # no threshold exceeded and trigger_on_false is False
                return None

        if time.time() - self.last_trigger < self.params.threshold.min_delay.value:
            # threshold exceeded but min_delay not reached
            return None

        if self.params.threshold.require_pass.value and not self.trigger_ready:
            # threshold exceeded but require_pass is True and threshold was already exceeded in the last cycle
            return None

        if np.any(thresholded):
            # threshold exceeded
            self.trigger_ready = False

        self.last_trigger = time.time()
        result = np.where(thresholded, self.params.threshold.true_value.value, self.params.threshold.false_value.value)
        self.triggered_last_cycle = True
        return {"thresholded": (result, data.meta)}



from goofi.data import Data, DataType
from goofi.node import Node
import numpy as np


class Coord2loc(Node):
    def config_input_slots():
        return {"latitude": DataType.ARRAY, "longitude": DataType.ARRAY}

    def config_output_slots():
        return {
            "coord_info": DataType.TABLE,
            "water_situation": DataType.ARRAY
        }

    def setup(self):
        from geopy.geocoders import Nominatim

        # list keys : ['ISO3166-2-lvl4', 'ISO3166-2-lvl6', 'ISO3166-2-lvl7', 'city',
        # 'city_district', 'country', 'country_code', 'county', 'district', 'hamlet', 'locality',
        # 'municipality', 'postcode', 'region', 'road', 'state', 'state_district', 'suburb', 'town', 'village']
        # Use a user-defined user agent to comply with Nominatim's usage policy
        self.road = ["road"]
        self.village = ["village", "hamlet"]

        self.city = ["city", "town", "locality", "municipality", "city_district"]
        self.state = ["state", "region", "state_district"]
        self.country = ["country"]
        self.geolocator = Nominatim(user_agent="my_application")

    def process(self, latitude: Data, longitude: Data):
        if latitude is None or latitude.data is None or longitude is None or longitude.data is None:
            return None
        # check if the input is within the range of the latitude and longitude
        if not (-90 <= latitude.data <= 90 and -180 <= longitude.data <= 180):
            return {"coord_info": ({"info : ": Data(DataType.STRING, "Invalid coordinates", {})}, {})}

        latitude = latitude.data
        longitude = longitude.data

        # Convert the input coordinates to a string format "latitude, longitude"
        coordinates = (latitude, longitude)
        # Use geolocator to find the location from coordinates
        location = self.geolocator.reverse(coordinates)

        # Check if location is found
        if location and location.address:

            location_info = {}
            for key in self.city:
                value = location.raw.get("address", {}).get(key)
                if value:
                    location_info["city"] = Data(DataType.STRING, value, {})
                    break
            value = None
            for key in self.state:
                value = location.raw.get("address", {}).get(key)
                if value:
                    location_info["state"] = Data(DataType.STRING, value, {})
                    break
            value = None
            for key in self.country:
                value = location.raw.get("address", {}).get(key)
                if value:
                    location_info["country"] = Data(DataType.STRING, value, {})
                    break
            value = None
            for key in self.road:
                value = location.raw.get("address", {}).get(key)
                if value:
                    location_info["road"] = Data(DataType.STRING, value, {})
                    break
            value = None
            for key in self.village:
                value = location.raw.get("address", {}).get(key)
                if value:
                    location_info["village"] = Data(DataType.STRING, value, {})
                    break
            location_info["full_address"] = Data(DataType.STRING, location.address, {})
            value = None

            return {"coord_info": (location_info, {}), "water_situation": (np.array(0), {})}
        else:
            return {"coord_info": ({"info": Data(DataType.STRING, "You're lost in the ocean", {})}, {}), "water_situation": (np.array(1), {})}



from multiprocessing import Manager as MPManager

import pytest

from goofi.connection import Connection
from goofi.data import Data, DataType
from goofi.message import Message, MessageType

try:
    mp_manager = MPManager()
    Connection.set_backend("mp", mp_manager)
except AssertionError:
    # connection backend is already set
    pass

EXAMPLE_CONTENT = {
    MessageType.ADD_OUTPUT_PIPE: {
        "slot_name_out": "out",
        "slot_name_in": "in",
        "node_connection": Connection.create()[0],
    },
    MessageType.REMOVE_OUTPUT_PIPE: {
        "slot_name_out": "out",
        "slot_name_in": "in",
        "node_connection": Connection.create()[0],
    },
    MessageType.DATA: {"slot_name": "test", "data": Data(DataType.STRING, "", {})},
    MessageType.CLEAR_DATA: {"slot_name": "test"},
    MessageType.PING: {},
    MessageType.PONG: {},
    MessageType.TERMINATE: {},
    MessageType.PROCESSING_ERROR: {"error": "test"},
    MessageType.PARAMETER_UPDATE: {"group": "common", "param_name": "test", "param_value": "test"},
    MessageType.SERIALIZE_REQUEST: {},
    MessageType.SERIALIZE_RESPONSE: {"_type": "test", "category": "test", "out_conns": {}, "params": {}},
}


@pytest.mark.parametrize("type", MessageType.__members__.values())
def test_create_message(type):
    if type not in EXAMPLE_CONTENT:
        # EXAMPLE_CONTENT is missing a test for this type
        raise NotImplementedError(f"Missing test for {type}.")

    # all type checks should pass
    Message(type, EXAMPLE_CONTENT[type])


@pytest.mark.parametrize("type", MessageType.__members__.values())
def test_message_content(type):
    if type not in EXAMPLE_CONTENT:
        # EXAMPLE_CONTENT is missing a test for this type
        raise NotImplementedError(f"Missing test for {type}.")

    for key in EXAMPLE_CONTENT[type]:
        # remove the key from the content
        content = EXAMPLE_CONTENT[type].copy()
        del content[key]

        # should raise ValueError
        with pytest.raises(ValueError):
            Message(type, content)


@pytest.mark.parametrize("type", MessageType.__members__.values())
def test_message_errors(type):
    if type not in EXAMPLE_CONTENT:
        # EXAMPLE_CONTENT is missing a test for this type
        raise NotImplementedError(f"Missing test for {type}.")

    # type is None
    with pytest.raises(ValueError):
        Message(None, EXAMPLE_CONTENT[type])
    # content is None
    with pytest.raises(ValueError):
        Message(type, None)



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class AudioOut(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"finished": DataType.ARRAY}

    def config_params():
        return {
            "audio": {
                "sampling_rate": StringParam("44100", options=["44100", "48000", "32000", "16000"]),
                "device": StringParam(AudioOut.list_audio_devices()[0], options=AudioOut.list_audio_devices()),
                "transition_samples": 100,
            }
        }

    def setup(self):
        import sounddevice as sd

        if hasattr(self, "stream") and self.stream:
            self.stream.stop()
            self.stream.close()

        self.stream = sd.OutputStream(
            samplerate=int(self.params.audio.sampling_rate.value),
            device=self.params.audio.device.value,
        )
        self.stream.start()

        self.last_sample = None

    def process(self, data: Data):
        if data is None:
            return

        if self.stream is None:
            raise RuntimeError("Audio output stream is not available.")

        # set data type to float32
        samples = data.data.astype(np.float32).T
        # Handle Mono to Stereo or Stereo to Mono Conversion
        # Verify that the samples array has the correct number of dimensions
        if samples.data.ndim == 1:
            # Mono audio: duplicate the channel for stereo output
            samples = np.stack((samples.data, samples.data), axis=-1)
        elif samples.data.ndim == 2 and samples.data.shape[1] == 1:
            # Also handle the case where the array is 2D but has only one channel
            samples = np.concatenate((samples.data, samples.data), axis=1)
        else:
            # For already stereo or multi-channel data, use as is
            samples = samples.data

        if self.last_sample is None:
            self.last_sample = samples[-1]

        transition = np.linspace(self.last_sample, samples[0], num=self.params.audio.transition_samples.value)
        samples = np.concatenate((transition, samples), axis=0)

        self.last_sample = samples[-1]

        # Send the audio data to the output device after ensuring it's C-contiguous
        self.stream.write(np.ascontiguousarray(samples))

        return {"finished": (np.array([1]), {})}

    def audio_sampling_rate_changed(self, value):
        self.setup()

    @staticmethod
    def list_audio_devices():
        """Returns a list of available audio devices."""
        import sounddevice as sd

        if sd is None:
            return ["None"]

        devices = sd.query_devices()
        device_names = []
        for device in devices:
            # check if the device is an output device
            if device["max_output_channels"] > 0:
                device_names.append(device["name"])
        return device_names

    def audio_device_changed(self, value):
        self.setup()



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam, StringParam


class SpikingNetwork(Node):
    def config_params():
        return {
            "network": {
                "n_neurons": IntParam(100, 10, 10000),
                "max_n_in": IntParam(20, 1, 100),
                "spatial_dims": IntParam(2, 1, 3),
                "device": StringParam("auto", options=["auto", "cpu", "cuda"]),
                "reset": BoolParam(False, trigger=True),
            },
            "simulation": {
                "ap_speed": FloatParam(0.1, 0.01, 1.0),
                "ap_threshold": FloatParam(1.0, 0.1, 5.0),
                "refraction_potential": FloatParam(-1.0, -10.0, 0.0),
                "potential_decay": FloatParam(0.1, 0.01, 1.0),
                "neurotransmitter_replenishment_rate": FloatParam(0.1, 0.01, 1.0),
                "weight_shift": FloatParam(0, -1.0, 1.0),
                "weight_scale": FloatParam(1.0, 0.1, 2.0),
                "delta_time": FloatParam(0.01, 0.01, 1.0),
            },
        }

    def config_input_slots():
        return {"input": DataType.ARRAY}

    def config_output_slots():
        return {"potentials": DataType.ARRAY}

    def setup(self):
        import torch

        self.torch = torch

        device = self.params.network.device.value
        if device == "auto":
            device = "cuda" if self.torch.cuda.is_available() else "cpu"

        # create the SNN instance with parameters
        self.network = Network(
            n_neurons=self.params.network.n_neurons.value,
            max_n_in=self.params.network.max_n_in.value,
            spatial_dims=self.params.network.spatial_dims.value,
            device=device,
        )

    def process(self, input: Data):
        # update the simulation parameters
        self.network.AP_SPEED = self.params.simulation.ap_speed.value
        self.network.AP_THRESHOLD = self.params.simulation.ap_threshold.value
        self.network.REFRACTION_POTENTIAL = self.params.simulation.refraction_potential.value
        self.network.POTENTIAL_DECAY = self.params.simulation.potential_decay.value
        self.network.NEUROTRANSMITTER_REPLENISHMENT_RATE = self.params.simulation.neurotransmitter_replenishment_rate.value

        if input is not None:
            assert input.data.ndim == 1, "Input must be one-dimensional."
            assert input.data.shape[0] < self.network.potential.shape[0], "Input size must match number of neurons."
            self.network.potential[: input.data.shape[0]] += self.torch.from_numpy(input.data).to(self.network.potential.device)

        # simulate the SNN for one time step
        self.network.update(
            dt=self.params.simulation.delta_time.value,
            shift=self.params.simulation.weight_shift.value,
            scale=self.params.simulation.weight_scale.value,
        )
        return {"potentials": (self.network.potential.cpu().numpy(), {"sfreq": 1 / self.params.simulation.delta_time.value})}

    def network_n_neurons_changed(self, value):
        self.setup()

    def network_max_n_in_changed(self, value):
        self.setup()

    def network_spatial_dims_changed(self, value):
        self.setup()

    def network_device_changed(self, value):
        self.setup()

    def network_reset_changed(self, value):
        if value:
            self.setup()


class Network:
    AP_SPEED = 0.1
    AP_THRESHOLD = 1.0
    REFRACTION_POTENTIAL = -1.0
    POTENTIAL_DECAY = 0.1
    NEUROTRANSMITTER_REPLENISHMENT_RATE = 0.1

    def __init__(self, n_neurons=5000, max_n_in=20, spatial_dims=2, device="cpu"):
        import torch

        self.torch = torch

        self.positions = self.torch.rand(n_neurons, spatial_dims, device=device)
        self.potential = self.torch.zeros(n_neurons, device=device)
        self.ap_timers = self.torch.full((n_neurons, max_n_in), float("inf"), device=device)

        self.dendrite_weights = self.torch.full((n_neurons, max_n_in), 0.0, device=device)
        self.dendrite_indices = self.torch.full((n_neurons, max_n_in), -1, device=device, dtype=self.torch.long)
        self.dendrite_activity = self.torch.ones(n_neurons, max_n_in, device=device)

        self.distances = self.torch.cdist(self.positions, self.positions).fill_diagonal_(float("inf"))

        for i in range(n_neurons):
            n = self.torch.randint(1, max_n_in, (1,)).item()
            self.dendrite_indices[i, :n] = self.torch.multinomial((1 / (self.distances[i] + 1e-6)).softmax(dim=0), n)
            self.dendrite_weights[i, :n] = self.torch.randn(n)

    def update(self, dt=0.1, shift=0, scale=1):
        self.ap_timers -= self.AP_SPEED * dt
        arrived = self.ap_timers < 0
        self.ap_timers[arrived] = float("inf")

        mask = self.potential.abs() < self.POTENTIAL_DECAY * dt
        self.potential[mask] = 0

        mask = ~mask
        if mask.sum() > 0:
            self.potential[mask] -= self.torch.sign(self.potential[mask]) * self.POTENTIAL_DECAY * dt

        self.potential += (arrived * (self.dendrite_weights * scale + shift) / self.dendrite_activity).sum(dim=-1)

        self.dendrite_activity = (self.dendrite_activity + arrived - self.NEUROTRANSMITTER_REPLENISHMENT_RATE * dt).clamp_min_(
            1
        )

        new_ap = self.potential >= self.AP_THRESHOLD
        self.potential[new_ap] = self.REFRACTION_POTENTIAL

        firing_neurons = new_ap.nonzero(as_tuple=True)[0]
        mask = self.torch.isin(self.dendrite_indices, firing_neurons) & self.ap_timers.isinf()
        receiving_neurons, dendrite_indices = mask.nonzero(as_tuple=True)
        self.ap_timers[receiving_neurons, dendrite_indices] = self.distances[
            receiving_neurons,
            self.dendrite_indices[receiving_neurons, dendrite_indices],
        ]



import numpy as np
from numpy.fft import fft, fftfreq

# from mne.time_frequency import tfr_array_multitaper
from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class PSD(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"psd": DataType.ARRAY}

    def config_params():
        return {
            "psd": {
                "method": StringParam("welch", options=["fft", "welch"]),
                "noverlap": IntParam(0, 0, 500),
                "precision": FloatParam(0.1, 0.01, 10.0),
                "f_min": FloatParam(1.0, 0.01, 9999.0),
                "f_max": FloatParam(60.0, 1.0, 10000.0),
                "smooth_welch": IntParam(1, 1, 10),
                # "time_bandwidth_multitaper": FloatParam(2.0, 0.1, 10.0),
                # "n_cycles_multitaper": IntParam(7, 1, 20),
            }
        }

    def setup(self):
        from scipy.signal import welch

        self.welch = welch

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        if data.data.ndim not in [1, 2]:
            raise ValueError("Data must be 1D or 2D")

        method = self.params["psd"]["method"].value
        noverlap = self.params["psd"]["noverlap"].value
        precision = self.params["psd"]["precision"].value
        f_min = self.params["psd"]["f_min"].value  # Get the min frequency
        f_max = self.params["psd"]["f_max"].value  # Get the max frequency
        smooth = self.params["psd"]["smooth_welch"].value
        # time_bandwidth = self.params["psd"]["time_bandwidth_multitaper"].value
        # n_cycles = self.params["psd"]["n_cycles_multitaper"].value

        sfreq = data.meta["sfreq"]
        nperseg = int(sfreq / precision)
        nfft = nperseg / smooth

        # Sanity check for recommended parameters
        if data.data.shape[-1] < f_min * 3 * sfreq:
            print(
                "Warning: The minimum frequency is too low for the length of the signal. "
                "Consider increasing the minimum frequency or increasing the signal length."
            )

        if method == "fft":
            freq = fftfreq(data.data.shape[-1], 1 / sfreq)
            fft_result = fft(data.data, axis=-1)
            psd = np.abs(fft_result)
        elif method == "welch":
            if data.data.ndim == 1:
                freq, psd = self.welch(data.data, fs=sfreq, nperseg=nperseg, nfft=nfft, noverlap=noverlap)
            else:  # if 2D
                psd = []
                for row in data.data:
                    f, p = self.welch(row, fs=sfreq, nperseg=nperseg, nfft=nfft, noverlap=noverlap)
                    psd.append(p)
                freq = f
                psd = np.array(psd)
        """elif method == "multitaper":
            fmin, fmax = f_min, f_max
            # Computing the TFR using multitaper
            power_tfr = tfr_array_multitaper(
                data.data[None],
                sfreq=sfreq,
                freqs=np.arange(fmin, fmax, precision),
                n_cycles=n_cycles,  # This can be adjusted or parameterized
                time_bandwidth=time_bandwidth,
                use_fft=True,
                verbose=False,
            )

            # Extract the PSD data
            # Sum the power across the time dimension to get PSD
            psd = np.mean(power_tfr.data, axis=-1)

            # The first dimension in the result is redundant and needs to be removed
            psd = psd.squeeze()
            freq = np.arange(fmin, fmax, precision)"""

        # prepare metadata
        meta = data.meta.copy()

        # Selecting the range of frequencies
        valid_indices = np.where((freq >= f_min) & (freq <= f_max))[0]
        freq = freq[valid_indices]
        if data.data.ndim == 1:
            psd = psd[valid_indices]
            meta["channels"]["dim0"] = freq.tolist()
        else:  # if 2D
            psd = psd[:, valid_indices]
            meta["channels"]["dim1"] = freq.tolist()

        return {"psd": (psd, meta)}



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam


class EMD(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"IMFs": DataType.ARRAY}

    def config_params():
        return {"emd": {"nIMFs": IntParam(5, 1, 10)}}

    def setup(self):
        from biotuner.peaks_extraction import EMD_eeg

        self.EMD_eeg = EMD_eeg

    def process(self, data: Data):
        if data is None:
            return None

        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        n_imfs = self.params.emd.nIMFs.value

        # add indices for each IMF in the meta data as strings
        imfs = self.EMD_eeg(data.data, method="EMD_fast", graph=False, extrema_detection="simple", nIMFs=5)[:n_imfs]
        data.meta["channels"]["dim0"] = ["IMF" + str(i) for i in range(imfs.shape[0])]

        return {"IMFs": (imfs, data.meta)}



import json
from os import environ, path

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam, StringParam


class TextGeneration(Node):

    @staticmethod
    def config_input_slots():
        return {"prompt": DataType.STRING}

    @staticmethod
    def config_output_slots():
        return {"generated_text": DataType.STRING}

    @staticmethod
    def config_params():
        return {
            "text_generation": {
                "api_key": StringParam("openai.key", doc="API key for the text generation service"),
                "model": StringParam("gpt-4o-mini", doc="Model ID for the text generation service"),
                "system_prompt": StringParam("", doc="System prompt for text generation"),
                "temperature": FloatParam(1.0, 0.0, 2.0, doc="Temperature for text generation"),
                "max_tokens": IntParam(20, 5, 2048, doc="Maximum number of tokens to generate"),
                "keep_conversation": BoolParam(False, doc="Whether to keep conversation history"),
                "save_conversation": StringParam("", doc="Whether to save conversation history to a JSON file"),
            }
        }

    def setup(self):
        import requests

        self.requests = requests

        self.client = None
        self.previous_model = None
        self.api_key_loaded = False
        self.messages = []
        self.load_api_key()
        self.import_libs(self.params["text_generation"]["model"].value)
        self.system_prompt = self.params["text_generation"]["system_prompt"].value

    def load_api_key(self):
        self.api_key = self.params["text_generation"]["api_key"].value
        model = self.params["text_generation"]["model"].value

        if path.exists(self.api_key):
            with open(self.api_key, "r") as f:
                self.api_key = f.read().strip()
        elif model.startswith("gpt-"):
            self.api_key = environ.get("OPENAI_API_KEY", self.api_key)
        elif model.startswith("claude-"):
            self.api_key = environ.get("ANTHROPIC_API_KEY", self.api_key)
        elif model.startswith("gemini-"):
            self.api_key = environ.get("GOOGLE_API_KEY", self.api_key)

        if not self.api_key and not model.startswith("local-"):
            raise ValueError(f"API key for {model} not found in environment variables or input parameters.")

        print(f"Loaded API key for {model}.")

    def import_libs(self, model):
        if model != self.previous_model:
            self.openai, self.anthropic, self.genai, self.ollama = self.dynamic_import_libs(model)  # Include ollama
            self.previous_model = model
            self.client = None

    def dynamic_import_libs(self, model):
        if model.startswith("gpt-"):
            try:
                import openai
            except ImportError:
                raise ImportError("You need to install openai: pip install openai")
            return openai, None, None, None
        elif model.startswith("claude-"):
            try:
                import anthropic
            except ImportError:
                raise ImportError("You need to install anthropic: pip install anthropic")
            return None, anthropic, None, None
        elif model.startswith("gemini-"):
            try:
                import google.generativeai as genai
            except ImportError:
                raise ImportError("You need to install google-generativeai: pip install google-generativeai\nWith Python>=3.10")
            return None, None, genai, None
        elif model.startswith("ollama-"):
            try:
                import ollama
            except ImportError:
                raise ImportError("You need to install ollama: pip install ollama")
            return None, None, None, ollama  # Return ollama
        else:
            raise ValueError(f"Unknown model: {model}")

    def api_key_changed(self):
        self.load_api_key()
        self.client = None
        self.api_key_loaded = True

    def generate_openai_response(self, messages, temp):
        if self.client is None:
            self.client = self.openai.OpenAI(api_key=self.api_key)

        payload = {
            "model": self.params["text_generation"]["model"].value,
            "messages": messages,
            "temperature": temp,
            "max_tokens": self.params["text_generation"]["max_tokens"].value,
        }

        if self.system_prompt is not None and len(messages) < 2:
            payload["messages"] = [{"role": "system", "content": self.system_prompt}] + messages

        response = self.client.chat.completions.create(**payload)
        return response.choices[0].message.content

    def generate_anthropic_response(self, messages, temp):
        if self.client is None:
            self.client = self.anthropic.Anthropic(api_key=self.api_key)

        response = self.client.messages.create(
            model=self.params["text_generation"]["model"].value,
            messages=messages,
            max_tokens=self.params["text_generation"]["max_tokens"].value,
            temperature=temp,
            system=self.system_prompt if self.system_prompt is not None else None,
        )

        return response.content[0].text

    def generate_gemini_response(self, text, temp, keep_conversation, history=None):
        if self.client is None:
            self.genai.configure(api_key=self.api_key)
            # If system prompt is provided, prepend it to the message
            self.client = self.genai.GenerativeModel(
                model_name=self.params["text_generation"]["model"].value,
                system_instruction=self.system_prompt if self.system_prompt is not None else "You are a helpful assistant.",
            )

        if not keep_conversation:
            history = []

        chat = self.client.start_chat(history=history)

        generation_config = self.genai.GenerationConfig(
            max_output_tokens=self.params["text_generation"]["max_tokens"].value, temperature=temp
        )

        # Prepare the message
        message = text

        # Send the request
        response = chat.send_message(message, generation_config=generation_config)
        # NOTE:  double check if history works this way
        try:
            history = chat.history()
        except Exception as e:
            print(f"History is None: {e}")
            history = None

        return response.text

    def generate_ollama_response(self, model, messages, temp):
        # Create an Ollama client instance
        if self.client is None:
            self.client = self.ollama

        response = self.client.chat(
            model=str(model.replace("ollama-", "")),  # remove ollama- prefix
            messages=messages,
            options={
                "system": self.system_prompt,
                "temperature": temp,
                "max_tokens": self.params["text_generation"]["max_tokens"].value,
            },
        )
        return response["message"]["content"]

    def generate_local_response(self, content):
        url = "http://127.0.0.1:5000/v1/chat/completions"
        headers = {"Content-Type": "application/json"}
        data = {
            "mode": "chat-instruct",  # TODO: Add mode to params
            "max_tokens": self.params["text_generation"]["max_tokens"].value,
            "temperature": self.params["text_generation"]["temperature"].value,
            "character": "Example",
            "messages": [{"role": "user", "content": content}],
        }
        response = self.requests.post(url, headers=headers, json=data, verify=False)
        return response.json()["choices"][0]["message"]["content"]

    def save_conversation_to_json(self):
        filename = f"conversation_{self.params['text_generation']['save_conversation'].value}.json"
        with open(filename, "w") as f:
            json.dump(self.messages, f, indent=4)
        print(f"Conversation saved to {filename}")

    def process(self, prompt: Data):
        if prompt is None or prompt.data is None:
            return None

        self.import_libs(self.params["text_generation"]["model"].value)

        prompt_ = prompt.data
        temp = self.params["text_generation"]["temperature"].value
        model = self.params["text_generation"]["model"].value
        keep_conversation = self.params["text_generation"]["keep_conversation"].value
        save_conversation = self.params["text_generation"]["save_conversation"].value

        if keep_conversation:
            self.messages.append({"role": "user", "content": prompt_})
        else:
            self.messages = [{"role": "user", "content": prompt_}]

        if model.startswith("gpt-"):
            if self.api_key_loaded:
                pass
            else:
                self.load_api_key()  # Ensure API key is loaded with the current model
            generated_text = self.generate_openai_response(self.messages, temp)

        elif model.startswith("claude-"):
            if self.api_key_loaded:
                pass
            else:
                self.load_api_key()
            generated_text = self.generate_anthropic_response(self.messages, temp)

        elif model.startswith("gemini-"):
            if self.api_key_loaded:
                pass
            else:
                self.load_api_key()
            generated_text = self.generate_gemini_response(prompt_, temp, keep_conversation)

        elif model.startswith("local-"):
            generated_text = self.generate_local_response(prompt_)

        elif model.startswith("ollama-"):
            generated_text = self.generate_ollama_response(model, self.messages, temp)
        else:
            raise ValueError(f"Unknown model: {model}")

        if keep_conversation:
            self.messages.append({"role": "assistant", "content": generated_text})

        if save_conversation:
            self.save_conversation_to_json()

        if self.params["text_generation"]["api_key"].value != self.api_key:
            self.api_key_changed()

        return {"generated_text": (generated_text, prompt.meta)}



from typing import Any, Dict, Tuple

import cv2
import numpy as np
from mss import mss

from goofi.data import DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam, StringParam


class VideoStream(Node):
    def config_params():
        return {
            "video_stream": {
                "capture_mode": StringParam("camera", ["camera", "screen"]),
                "device_index": IntParam(0, 0, 10, doc="Index of the camera or screen to capture"),
                "mirror": BoolParam(False),
                "crop_left": IntParam(0, 0, 2000, doc="Number of pixels to crop from the left"),
                "crop_top": IntParam(0, 0, 2000, doc="Number of pixels to crop from the top"),
                "crop_right": IntParam(0, 0, 2000, doc="Number of pixels to crop from the right"),
                "crop_bottom": IntParam(0, 0, 2000, doc="Number of pixels to crop from the bottom"),
            },
            "common": {"autotrigger": BoolParam(True)},
        }

    def config_output_slots():
        return {"frame": DataType.ARRAY}

    def setup(self):
        try:
            # close opencv video capture
            self.cap.release()
        except:
            pass

        self.cap = None
        if self.params.video_stream.capture_mode.value == "camera":
            self.cap = cv2.VideoCapture(self.params.video_stream.device_index.value)

    def process(self) -> Dict[str, Tuple[Any, Dict[str, Any]]]:
        frame = None

        capture_mode = self.params.video_stream.capture_mode.value
        cropping_done = False
        if capture_mode == "camera":
            if self.cap is None:
                return None

            ret, frame = self.cap.read()
            if ret:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            else:
                print("Failed to capture frame from camera.")
                return None

        elif capture_mode == "screen":
            screen_index = self.params.video_stream.device_index.value

            with mss(display=screen_index) as sct:
                if screen_index >= len(sct.monitors):
                    print(f"Screen index {screen_index} is out of range.")

                # update monitor size with cropping values
                monitor_size = sct.monitors[screen_index].copy()
                monitor_size["top"] += self.params.video_stream.crop_top.value
                monitor_size["left"] += self.params.video_stream.crop_left.value
                monitor_size["width"] -= self.params.video_stream.crop_left.value + self.params.video_stream.crop_right.value
                monitor_size["height"] -= self.params.video_stream.crop_top.value + self.params.video_stream.crop_bottom.value

                # cropping is done by mss to avoid performance overhead
                cropping_done = True
                screen_capture = sct.grab(monitor_size)

                frame = np.array(screen_capture)
                frame = cv2.cvtColor(frame, cv2.COLOR_BGRA2RGB)

        if frame is None:
            return None

        if not cropping_done:
            # crop the frame if still needed
            frame = frame[
                self.params.video_stream.crop_top.value : -self.params.video_stream.crop_bottom.value - 1,
                self.params.video_stream.crop_left.value : -self.params.video_stream.crop_right.value - 1,
            ]

        if self.params.video_stream.mirror.value:
            # flip the frame horizontally
            frame = cv2.flip(frame, 1)

        frame = frame.astype("float32") / 255.0

        return {"frame": (frame, {})}

    def video_stream_device_index_changed(self, value):
        # opencv video capture needs to be reinitialized, mss can be reused
        if self.params.video_stream.capture_mode.value == "camera":
            self.setup()

    def video_stream_capture_mode_changed(self, value):
        self.setup()
